@@ -4,11 +4,11 @@
 #include <stdbool.h>
 #include <uchar.h>
 
+#include <xkbcommon/xkbcommon.h>
 #include <tllist.h>
+#include <fcft/fcft.h>
 
-#include "terminal.h"
 #include "user-notification.h"
-#include "wayland.h"
 
 #define DEFINE_LIST(type) \
     type##_list {         \
@@ -16,6 +16,14 @@
         type *arr;        \
     }
 
+/* If px != 0 then px is valid, otherwise pt is valid */
+struct pt_or_px {
+    int16_t px;
+    float pt;
+};
+
+enum cursor_style { CURSOR_BLOCK, CURSOR_UNDERLINE, CURSOR_BEAM };
+
 enum conf_size_type {CONF_SIZE_PX, CONF_SIZE_CELLS};
 
 struct config_font {
@@ -36,16 +44,36 @@ struct argv {
     char **args;
 };
 
-struct config_binding_pipe {
-    struct argv argv;
-    bool master_copy;
+enum binding_aux_type {
+    BINDING_AUX_NONE,
+    BINDING_AUX_PIPE,
+    BINDING_AUX_TEXT,
 };
 
-enum config_key_binding_type {
+struct binding_aux {
+    enum binding_aux_type type;
+    bool master_copy;
+
+    union {
+        struct argv pipe;
+
+        struct {
+            uint8_t *data;
+            size_t len;
+        } text;
+    };
+};
+
+enum key_binding_type {
     KEY_BINDING,
     MOUSE_BINDING,
 };
 
+struct config_key_binding_text {
+    char *text;
+    bool master_copy;
+};
+
 struct config_key_binding {
     int action;  /* One of the varios bind_action_* enums from wayland.h */
     struct config_key_modifiers modifiers;
@@ -62,7 +90,7 @@ struct config_key_binding {
         } m;
     };
 
-    struct config_binding_pipe pipe;
+    struct binding_aux aux;
 
     /* For error messages in collision handling */
     const char *path;
@@ -76,6 +104,12 @@ struct config_spawn_template {
     struct argv argv;
 };
 
+struct env_var {
+    char *name;
+    char *value;
+};
+typedef tll(struct env_var) env_var_list_t;
+
 struct config {
     char *term;
     char *shell;
@@ -183,11 +217,25 @@ struct config {
             uint32_t bg;
         } scrollback_indicator;
 
+        struct {
+            struct {
+                uint32_t fg;
+                uint32_t bg;
+            } no_match;
+
+            struct {
+                uint32_t fg;
+                uint32_t bg;
+            } match;
+        } search_box;
+
         struct {
             bool selection:1;
             bool jump_label:1;
             bool scrollback_indicator:1;
             bool url:1;
+            bool search_box_no_match:1;
+            bool search_box_match:1;
             uint8_t dim;
         } use_custom;
     } colors;
@@ -234,6 +282,8 @@ struct config {
         uint16_t border_width_visible;
         uint16_t button_width;
 
+        bool hide_when_maximized;
+
         struct {
             bool title_set:1;
             bool buttons_set:1;
@@ -245,7 +295,7 @@ struct config {
             uint32_t buttons;
             uint32_t minimize;
             uint32_t maximize;
-            uint32_t close;
+            uint32_t quit;  /* ‘close’ collides with #define in epoll-shim */
             uint32_t border;
         } color;
 
@@ -266,6 +316,8 @@ struct config {
     struct config_spawn_template notify;
     bool notify_focus_inhibit;
 
+    env_var_list_t env_vars;
+
     struct {
         enum fcft_scaling_filter fcft_filter;
         bool overflowing_glyphs;
@@ -288,6 +340,7 @@ struct config {
         float box_drawing_base_thickness;
         bool box_drawing_solid_shades;
         bool font_monospace_warn;
+        bool sixel;
     } tweak;
 
     user_notifications_t notifications;
@@ -299,11 +352,16 @@ bool config_load(
     struct config *conf, const char *path,
     user_notifications_t *initial_user_notifications,
     config_override_t *overrides, bool errors_are_fatal);
-void config_free(struct config conf);
+void config_free(struct config *conf);
 struct config *config_clone(const struct config *old);
 
 bool config_font_parse(const char *pattern, struct config_font *font);
 void config_font_list_destroy(struct config_font_list *font_list);
 
+struct seat;
+xkb_mod_mask_t
+conf_modifiers_to_mask(
+    const struct seat *seat, const struct config_key_modifiers *modifiers);
+
 bool check_if_font_is_monospaced(
     const char *pattern, user_notifications_t *notifications);
