@@ -118,15 +118,19 @@ csi_sgr(struct terminal *term)
             term->vt.attrs.fg = param - 30;
             break;
 
-        case 38: {
+        case 38:
+        case 48: {
+            uint32_t color;
+            enum color_source src;
+
             /* Indexed: 38;5;<idx> */
             if (term->vt.params.idx - i - 1 >= 2 &&
                 term->vt.params.v[i + 1].value == 5)
             {
-                term->vt.attrs.fg_src = COLOR_BASE256;
-                term->vt.attrs.fg = term->vt.params.v[i + 2].value;
+                src = COLOR_BASE256;
+                color = min(term->vt.params.v[i + 2].value,
+                            ALEN(term->colors.table) - 1);
                 i += 2;
-
             }
 
             /* RGB: 38;2;<r>;<g>;<b> */
@@ -136,8 +140,8 @@ csi_sgr(struct terminal *term)
                 uint8_t r = term->vt.params.v[i + 2].value;
                 uint8_t g = term->vt.params.v[i + 3].value;
                 uint8_t b = term->vt.params.v[i + 4].value;
-                term->vt.attrs.fg_src = COLOR_RGB;
-                term->vt.attrs.fg = r << 16 | g << 8 | b;
+                src = COLOR_RGB;
+                color = r << 16 | g << 8 | b;
                 i += 4;
             }
 
@@ -145,10 +149,9 @@ csi_sgr(struct terminal *term)
             else if (term->vt.params.v[i].sub.idx >= 2 &&
                      term->vt.params.v[i].sub.value[0] == 5)
             {
-                const struct vt_param *param = &term->vt.params.v[i];
-
-                term->vt.attrs.fg_src = COLOR_BASE256;
-                term->vt.attrs.fg = param->sub.value[1];
+                src = COLOR_BASE256;
+                color = min(term->vt.params.v[i].sub.value[1],
+                            ALEN(term->colors.table) - 1);
             }
 
             /*
@@ -177,8 +180,8 @@ csi_sgr(struct terminal *term)
                 uint8_t g = param->sub.value[g_idx];
                 uint8_t b = param->sub.value[b_idx];
 
-                term->vt.attrs.fg_src = COLOR_RGB;
-                term->vt.attrs.fg = r << 16 | g << 8 | b;
+                src = COLOR_RGB;
+                color = r << 16 | g << 8 | b;
             }
 
             /* Transparent: 38:1 */
@@ -186,9 +189,19 @@ csi_sgr(struct terminal *term)
             /* CMYK:        38:4:<color-space>:c:m:y:k[:tolerance:tolerance-color-space] */
 
             /* Unrecognized */
-            else
+            else {
                 UNHANDLED_SGR(i);
+                break;
+            }
 
+            if (param == 38) {
+                term->vt.attrs.fg_src = src;
+                term->vt.attrs.fg = color;
+            } else {
+                xassert(param == 48);
+                term->vt.attrs.bg_src = src;
+                term->vt.attrs.bg = color;
+            }
             break;
         }
 
@@ -209,78 +222,6 @@ csi_sgr(struct terminal *term)
             term->vt.attrs.bg = param - 40;
             break;
 
-        case 48: {
-            /* Indexed: 48;5;<idx> */
-            if (term->vt.params.idx - i - 1 >= 2 &&
-                term->vt.params.v[i + 1].value == 5)
-            {
-                term->vt.attrs.bg_src = COLOR_BASE256;
-                term->vt.attrs.bg = term->vt.params.v[i + 2].value;
-                i += 2;
-
-            }
-
-            /* RGB: 48;2;<r>;<g>;<b> */
-            else if (term->vt.params.idx - i - 1 >= 4 &&
-                     term->vt.params.v[i + 1].value == 2)
-            {
-                uint8_t r = term->vt.params.v[i + 2].value;
-                uint8_t g = term->vt.params.v[i + 3].value;
-                uint8_t b = term->vt.params.v[i + 4].value;
-                term->vt.attrs.bg_src = COLOR_RGB;
-                term->vt.attrs.bg = r << 16 | g << 8 | b;
-                i += 4;
-            }
-
-            /* Indexed: 48:5:<idx> */
-            else if (term->vt.params.v[i].sub.idx >= 2 &&
-                     term->vt.params.v[i].sub.value[0] == 5)
-            {
-                const struct vt_param *param = &term->vt.params.v[i];
-
-                term->vt.attrs.bg_src = COLOR_BASE256;
-                term->vt.attrs.bg = param->sub.value[1];
-            }
-
-            /*
-             * RGB: 48:2:<color-space>:r:g:b[:ignored:tolerance:tolerance-color-space]
-             * RGB: 48:2:r:g:b
-             *
-             * The second version is a "bastard" version - many
-             * programs "forget" the color space ID
-             * parameter... *sigh*
-             */
-            else if (term->vt.params.v[i].sub.idx >= 4 &&
-                     term->vt.params.v[i].sub.value[0] == 2)
-            {
-                const struct vt_param *param = &term->vt.params.v[i];
-                bool have_color_space_id = param->sub.idx >= 5;
-
-                /* 0 - color space (ignored) */
-                int r_idx = 2 - !have_color_space_id;
-                int g_idx = 3 - !have_color_space_id;
-                int b_idx = 4 - !have_color_space_id;
-                /* 5 - unused */
-                /* 6 - CS tolerance */
-                /* 7 - color space associated with tolerance */
-
-                uint8_t r = param->sub.value[r_idx];
-                uint8_t g = param->sub.value[g_idx];
-                uint8_t b = param->sub.value[b_idx];
-
-                term->vt.attrs.bg_src = COLOR_RGB;
-                term->vt.attrs.bg = r << 16 | g << 8 | b;
-            }
-
-            /* Transparent: 48:1 */
-            /* CMY:         48:3:<color-space>:c:m:y[:tolerance:tolerance-color-space] */
-            /* CMYK:        48:4:<color-space>:c:m:y:k[:tolerance:tolerance-color-space] */
-
-            else
-                UNHANDLED_SGR(i);
-
-            break;
-        }
         case 49:
             term->vt.attrs.bg_src = COLOR_DEFAULT;
             break;
@@ -393,6 +334,11 @@ decset_decrst(struct terminal *term, unsigned param, bool enable)
         term->reverse_wrap = enable;
         break;
 
+    case 66:
+        /* DECNKM */
+        term->keypad_keys_mode = enable ? KEYPAD_APPLICATION : KEYPAD_NUMERICAL;
+        break;
+
     case 80:
         term->sixel.scrolling = !enable;
         break;
@@ -564,10 +510,6 @@ decset_decrst(struct terminal *term, unsigned param, bool enable)
         term->sixel.cursor_right_of_graphics = enable;
         break;
 
-    case 27127:
-        term->modify_escape_key = enable;
-        break;
-
     case 737769:
         if (enable)
             term_ime_enable(term);
@@ -607,6 +549,7 @@ decrqm(const struct terminal *term, unsigned param, bool *enabled)
     case 12: *enabled = term->cursor_blink.decset; return true;
     case 25: *enabled = !term->hide_cursor; return true;
     case 45: *enabled = term->reverse_wrap; return true;
+    case 66: *enabled = term->keypad_keys_mode == KEYPAD_APPLICATION; return true;
     case 80: *enabled = !term->sixel.scrolling; return true;
     case 1000: *enabled = term->mouse_tracking == MOUSE_CLICK; return true;
     case 1001: *enabled = false; return true;
@@ -625,11 +568,10 @@ decrqm(const struct terminal *term, unsigned param, bool *enabled)
     case 47:   /* FALLTHROUGH */
     case 1047: /* FALLTHROUGH */
     case 1049: *enabled = term->grid == &term->alt; return true;
-    case 1079: *enabled = term->sixel.use_private_palette; return true;
+    case 1070: *enabled = term->sixel.use_private_palette; return true;
     case 2004: *enabled = term->bracketed_paste; return true;
     case 2026: *enabled = term->render.app_sync_updates.enabled; return true;
     case 8452: *enabled = term->sixel.cursor_right_of_graphics; return true;
-    case 27127: *enabled = term->modify_escape_key; return true;
     case 737769: *enabled = term_ime_is_enabled(term); return true;
     }
 
@@ -651,6 +593,7 @@ xtsave(struct terminal *term, unsigned param)
     case 25: term->xtsave.show_cursor = !term->hide_cursor; break;
     case 45: term->xtsave.reverse_wrap = term->reverse_wrap; break;
     case 47: term->xtsave.alt_screen = term->grid == &term->alt; break;
+    case 66: term->xtsave.application_keypad_keys = term->keypad_keys_mode == KEYPAD_APPLICATION; break;
     case 80: term->xtsave.sixel_display_mode = !term->sixel.scrolling; break;
     case 1000: term->xtsave.mouse_click = term->mouse_tracking == MOUSE_CLICK; break;
     case 1001: break;
@@ -673,7 +616,6 @@ xtsave(struct terminal *term, unsigned param)
     case 2004: term->xtsave.bracketed_paste = term->bracketed_paste; break;
     case 2026: term->xtsave.app_sync_updates = term->render.app_sync_updates.enabled; break;
     case 8452: term->xtsave.sixel_cursor_right_of_graphics = term->sixel.cursor_right_of_graphics; break;
-    case 27127: term->xtsave.modify_escape_key = term->modify_escape_key; break;
     case 737769: term->xtsave.ime = term_ime_is_enabled(term); break;
     }
 }
@@ -694,6 +636,7 @@ xtrestore(struct terminal *term, unsigned param)
     case 25: enable = term->xtsave.show_cursor; break;
     case 45: enable = term->xtsave.reverse_wrap; break;
     case 47: enable = term->xtsave.alt_screen; break;
+    case 66: enable = term->xtsave.application_keypad_keys; break;
     case 80: enable = term->xtsave.sixel_display_mode; break;
     case 1000: enable = term->xtsave.mouse_click; break;
     case 1001: return;
@@ -716,7 +659,6 @@ xtrestore(struct terminal *term, unsigned param)
     case 2004: enable = term->xtsave.bracketed_paste; break;
     case 2026: enable = term->xtsave.app_sync_updates; break;
     case 8452: enable = term->xtsave.sixel_cursor_right_of_graphics; break;
-    case 27127: enable = term->xtsave.modify_escape_key; break;
     case 737769: enable = term->xtsave.ime; break;
 
     default: return;
@@ -794,8 +736,13 @@ csi_dispatch(struct terminal *term, uint8_t final)
              *
              * Note: tertiary DA responds with "FOOT".
              */
-            static const char reply[] = "\033[?62;4;22c";
-            term_to_slave(term, reply, sizeof(reply) - 1);
+            if (term->conf->tweak.sixel) {
+                static const char reply[] = "\033[?62;4;22c";
+                term_to_slave(term, reply, sizeof(reply) - 1);
+            } else {
+                static const char reply[] = "\033[?62;22c";
+                term_to_slave(term, reply, sizeof(reply) - 1);
+            }
             break;
         }
 
@@ -1417,6 +1364,11 @@ csi_dispatch(struct terminal *term, uint8_t final)
             break;
 
         case 'S': {
+            if (!term->conf->tweak.sixel) {
+                UNHANDLED();
+                break;
+            }
+
             unsigned target = vt_param_get(term, 0, 0);
             unsigned operation = vt_param_get(term, 1, 0);
 
@@ -1547,6 +1499,24 @@ csi_dispatch(struct terminal *term, uint8_t final)
             }
             break; /* final == 'm' */
 
+        case 'n': {
+            int resource = vt_param_get(term, 0, 2);  /* Default is modifyFuncionKeys */
+            switch (resource) {
+            case 0:  /* modifyKeyboard */
+            case 1:  /* modifyCursorKeys */
+            case 2:  /* modifyFunctionKeys */
+                break;
+
+            case 4:  /* modifyOtherKeys */
+                /* We don’t support fully disabling modifyOtherKeys,
+                 * but simply revert back to mode ‘1’ */
+                term->modify_other_keys_2 = false;
+                LOG_DBG("modifyOtherKeys=1");
+                break;
+            }
+            break;
+        }
+
         case 'u': {
             int flags = vt_param_get(term, 0, 0) & KITTY_KBD_SUPPORTED;
 
