@@ -10,28 +10,21 @@
 #include "vt.h"
 #include "xmalloc.h"
 
-static void
-bsu(struct terminal *term)
+static bool
+ensure_size(struct terminal *term, size_t required_size)
 {
-    /* https://gitlab.com/gnachman/iterm2/-/wikis/synchronized-updates-spec */
+    if (required_size <= term->vt.dcs.size)
+        return true;
 
-    size_t n = term->vt.dcs.idx;
-    if (unlikely(n > 0))
-        LOG_DBG("BSU with unknown params: %.*s)", (int)n, term->vt.dcs.data);
+    uint8_t *new_data = realloc(term->vt.dcs.data, required_size);
+    if (new_data == NULL) {
+        LOG_ERRNO("failed to increase size of DCS buffer");
+        return false;
+    }
 
-    term_enable_app_sync_updates(term);
-}
-
-static void
-esu(struct terminal *term)
-{
-    /* https://gitlab.com/gnachman/iterm2/-/wikis/synchronized-updates-spec */
-
-    size_t n = term->vt.dcs.idx;
-    if (unlikely(n > 0))
-        LOG_DBG("ESU with unknown params: %.*s)", (int)n, term->vt.dcs.data);
-
-    term_disable_app_sync_updates(term);
+    term->vt.dcs.data = new_data;
+    term->vt.dcs.size = required_size;
+    return true;
 }
 
 /* Decode hex-encoded string *inline*. NULL terminates */
@@ -184,6 +177,24 @@ out:
     free(name);
 }
 
+static void
+xtgettcap_put(struct terminal *term, uint8_t c)
+{
+    struct vt *vt = &term->vt;
+
+    /* Grow buffer expontentially */
+    if (vt->dcs.idx >= vt->dcs.size) {
+        size_t new_size = vt->dcs.size * 2;
+        if (new_size == 0)
+            new_size = 128;
+
+        if (!ensure_size(term, new_size))
+            return;
+    }
+
+    vt->dcs.data[vt->dcs.idx++] = c;
+}
+
 static void
 xtgettcap_unhook(struct terminal *term)
 {
@@ -224,7 +235,20 @@ append_sgr_attr_n(char **reply, size_t *len, const char *attr, size_t n)
 }
 
 static void
-decrqss(struct terminal *term)
+decrqss_put(struct terminal *term, uint8_t c)
+{
+    /* Largest request we support is two bytes */
+    if (!ensure_size(term, 2))
+        return;
+
+    struct vt *vt = &term->vt;
+    if (vt->dcs.idx > 2)
+        return;
+    vt->dcs.data[vt->dcs.idx++] = c;
+}
+
+static void
+decrqss_unhook(struct terminal *term)
 {
     const uint8_t *query = term->vt.dcs.data;
     const size_t n = term->vt.dcs.idx;
@@ -240,7 +264,7 @@ decrqss(struct terminal *term)
      * https://github.com/hackerb9/vt340test/issues/13
      */
 
-    if (memcmp(query, "r", n) == 0) {
+    if (n == 1 && query[0] == 'r') {
         /* DECSTBM - Set Top and Bottom Margins */
         char reply[64];
         int len = snprintf(reply, sizeof(reply), "\033P1$r%d;%dr\033\\",
@@ -249,7 +273,7 @@ decrqss(struct terminal *term)
         term_to_slave(term, reply, len);
     }
 
-    else if (memcmp(query, "m", n) == 0) {
+    else if (n == 1 && query[0] == 'm') {
         /* SGR - Set Graphic Rendition */
         char *reply = NULL;
         size_t len = 0;
@@ -356,7 +380,7 @@ decrqss(struct terminal *term)
         free(reply);
     }
 
-    else if (memcmp(query, " q", n) == 0) {
+    else if (n == 2 && memcmp(query, " q", 2) == 0) {
         /* DECSCUSR - Set Cursor Style */
         int mode;
 
@@ -376,7 +400,7 @@ decrqss(struct terminal *term)
     }
 
     else {
-        const char err[] = "\033P0$r\033\\";
+        static const char err[] = "\033P0$r\033\\";
         term_to_slave(term, err, sizeof(err) - 1);
     }
 }
@@ -396,6 +420,9 @@ dcs_hook(struct terminal *term, uint8_t final)
     case 0:
         switch (final) {
         case 'q': {
+            if (!term->conf->tweak.sixel) {
+                break;
+            }
             int p1 = vt_param_get(term, 0, 0);
             int p2 = vt_param_get(term, 1,0);
             int p3 = vt_param_get(term, 2, 0);
@@ -411,7 +438,8 @@ dcs_hook(struct terminal *term, uint8_t final)
     case '$':
         switch (final) {
         case 'q':
-            term->vt.dcs.unhook_handler = &decrqss;
+            term->vt.dcs.put_handler = &decrqss_put;
+            term->vt.dcs.unhook_handler = &decrqss_unhook;
             break;
         }
         break;
@@ -419,9 +447,14 @@ dcs_hook(struct terminal *term, uint8_t final)
     case '=':
         switch (final) {
         case 's':
+            /* BSU/ESU: https://gitlab.com/gnachman/iterm2/-/wikis/synchronized-updates-spec */
             switch (vt_param_get(term, 0, 0)) {
-            case 1: term->vt.dcs.unhook_handler = &bsu; return;
-            case 2: term->vt.dcs.unhook_handler = &esu; return;
+            case 1:
+                term->vt.dcs.unhook_handler = &term_enable_app_sync_updates;
+                return;
+            case 2:
+                term->vt.dcs.unhook_handler = &term_disable_app_sync_updates;
+                return;
             }
             break;
         }
@@ -430,6 +463,7 @@ dcs_hook(struct terminal *term, uint8_t final)
     case '+':
         switch (final) {
         case 'q':  /* XTGETTCAP */
+            term->vt.dcs.put_handler = &xtgettcap_put;
             term->vt.dcs.unhook_handler = &xtgettcap_unhook;
             break;
         }
@@ -437,26 +471,6 @@ dcs_hook(struct terminal *term, uint8_t final)
     }
 }
 
-static bool
-ensure_size(struct terminal *term, size_t required_size)
-{
-    if (required_size <= term->vt.dcs.size)
-        return true;
-
-    size_t new_size = (required_size + 127) / 128 * 128;
-    xassert(new_size > 0);
-
-    uint8_t *new_data = realloc(term->vt.dcs.data, new_size);
-    if (new_data == NULL) {
-        LOG_ERRNO("failed to increase size of DCS buffer");
-        return false;
-    }
-
-    term->vt.dcs.data = new_data;
-    term->vt.dcs.size = new_size;
-    return true;
-}
-
 void
 dcs_put(struct terminal *term, uint8_t c)
 {
@@ -464,11 +478,6 @@ dcs_put(struct terminal *term, uint8_t c)
 
     if (term->vt.dcs.put_handler != NULL)
         term->vt.dcs.put_handler(term, c);
-    else {
-        if (!ensure_size(term, term->vt.dcs.idx + 1))
-            return;
-        term->vt.dcs.data[term->vt.dcs.idx++] = c;
-    }
 }
 
 void
