@@ -15,6 +15,67 @@
 
 #define TIME_REFLOW 0
 
+/*
+ * “sb” (scrollback relative) coordinates
+ *
+ * The scrollback relative row number 0 is the *first*, and *oldest*
+ * row in the scrollback history (and thus the *first* row to be
+ * scrolled out). Thus, a higher number means further *down* in the
+ * scrollback, with the *highest* number being at the bottom of the
+ * screen, where new input appears.
+ */
+
+int
+grid_row_abs_to_sb(const struct grid *grid, int screen_rows, int abs_row)
+{
+    const int scrollback_start = grid->offset + screen_rows;
+    int rebased_row = abs_row - scrollback_start + grid->num_rows;
+
+    rebased_row &= grid->num_rows - 1;
+    return rebased_row;
+}
+
+int grid_row_sb_to_abs(const struct grid *grid, int screen_rows, int sb_rel_row)
+{
+    const int scrollback_start = grid->offset + screen_rows;
+    int abs_row = sb_rel_row + scrollback_start;
+
+    abs_row &= grid->num_rows - 1;
+    return abs_row;
+}
+
+int
+grid_sb_start_ignore_uninitialized(const struct grid *grid, int screen_rows)
+{
+    int scrollback_start = grid->offset + screen_rows;
+    scrollback_start &= grid->num_rows - 1;
+
+    while (grid->rows[scrollback_start] == NULL) {
+        scrollback_start++;
+        scrollback_start &= grid->num_rows - 1;
+    }
+
+    return scrollback_start;
+}
+
+int
+grid_row_abs_to_sb_precalc_sb_start(const struct grid *grid, int sb_start,
+                                    int abs_row)
+{
+    int rebased_row = abs_row - sb_start + grid->num_rows;
+    rebased_row &= grid->num_rows - 1;
+    return rebased_row;
+}
+
+int
+grid_row_sb_to_abs_precalc_sb_start(const struct grid *grid, int sb_start,
+                                    int sb_rel_row)
+{
+    int abs_row = sb_rel_row + sb_start;
+    abs_row &= grid->num_rows - 1;
+    return abs_row;
+}
+
 static void
 ensure_row_has_extra_data(struct row *row)
 {
@@ -168,6 +229,7 @@ grid_snapshot(const struct grid *grid)
         clone_row->cells = xmalloc(grid->num_cols * sizeof(clone_row->cells[0]));
         clone_row->linebreak = row->linebreak;
         clone_row->dirty = row->dirty;
+        clone_row->prompt_marker = row->prompt_marker;
 
         for (int c = 0; c < grid->num_cols; c++)
             clone_row->cells[c] = row->cells[c];
@@ -258,6 +320,7 @@ grid_row_alloc(int cols, bool initialize)
     row->dirty = false;
     row->linebreak = false;
     row->extra = NULL;
+    row->prompt_marker = false;
 
     if (initialize) {
         row->cells = xcalloc(cols, sizeof(row->cells[0]));
@@ -316,6 +379,7 @@ grid_resize_without_reflow(
 
         new_row->dirty = old_row->dirty;
         new_row->linebreak = false;
+        new_row->prompt_marker = old_row->prompt_marker;
 
         if (new_cols > old_cols) {
             /* Clear "new" columns */
@@ -475,6 +539,7 @@ _line_wrap(struct grid *old_grid, struct row **new_grid, struct row *row,
         /* Scrollback is full, need to re-use a row */
         grid_row_reset_extra(new_row);
         new_row->linebreak = false;
+        new_row->prompt_marker = false;
 
         tll_foreach(old_grid->sixel_images, it) {
             if (it->item.pos.row == *row_idx) {
@@ -482,6 +547,13 @@ _line_wrap(struct grid *old_grid, struct row **new_grid, struct row *row,
                 tll_remove(old_grid->sixel_images, it);
             }
         }
+
+        /*
+         * TODO: detect if the re-used row is covered by the
+         * selection. Of so, cancel the selection. The problem: we
+         * don’t know if we’ve translated the selection coordinates
+         * yet.
+         */
     }
 
     struct row_data *extra = row->extra;
@@ -668,6 +740,11 @@ grid_resize_and_reflow(
             }
         }
 
+        if (!old_row->linebreak && col_count > 0) {
+            /* Don’t truncate logical lines */
+            col_count = old_cols;
+        }
+
         xassert(col_count >= 0 && col_count <= old_cols);
 
         /* Do we have a (at least one) tracking point on this row */
@@ -794,6 +871,9 @@ grid_resize_and_reflow(
                 xassert(new_col_idx + amount <= new_cols);
                 xassert(from + amount <= old_cols);
 
+                if (from == 0)
+                    new_row->prompt_marker = old_row->prompt_marker;
+
                 memcpy(
                     &new_row->cells[new_col_idx], &old_row->cells[from],
                     amount * sizeof(struct cell));
@@ -855,13 +935,29 @@ grid_resize_and_reflow(
             start += cols;
         }
 
-
         if (old_row->linebreak) {
             /* Erase the remaining cells */
             memset(&new_row->cells[new_col_idx], 0,
                    (new_cols - new_col_idx) * sizeof(new_row->cells[0]));
             new_row->linebreak = true;
-            line_wrap();
+
+            if (r + 1 < old_rows)
+                line_wrap();
+            else if (new_row->extra != NULL &&
+                     new_row->extra->uri_ranges.count > 0)
+            {
+                /*
+                 * line_wrap() "closes" still-open URIs. Since this is
+                 * the *last* row, and since we’re line-breaking due
+                 * to a hard line-break (rather than running out of
+                 * cells in the "new_row"), there shouldn’t be an open
+                 * URI (it would have been closed when we reached the
+                 * end of the URI while reflowing the last "old"
+                 * row).
+                 */
+                uint32_t last_idx = new_row->extra->uri_ranges.count - 1;
+                xassert(new_row->extra->uri_ranges.v[last_idx].end >= 0);
+            }
         }
 
         grid_row_free(old_grid[old_row_idx]);
@@ -904,6 +1000,7 @@ grid_resize_and_reflow(
 
     /* Set offset such that the last reflowed row is at the bottom */
     grid->offset = new_row_idx - new_screen_rows + 1;
+
     while (grid->offset < 0)
         grid->offset += new_rows;
     while (new_grid[grid->offset] == NULL)
@@ -916,23 +1013,6 @@ grid_resize_and_reflow(
             new_grid[idx] = grid_row_alloc(new_cols, true);
     }
 
-    grid->view = view_follows ? grid->offset : viewport.row;
-
-    /* If enlarging the window, the old viewport may be too far down,
-     * with unallocated rows. Make sure this cannot happen */
-    while (true) {
-        int idx = (grid->view + new_screen_rows - 1) & (new_rows - 1);
-        if (new_grid[idx] != NULL)
-            break;
-        grid->view--;
-        if (grid->view < 0)
-            grid->view += new_rows;
-    }
-    for (size_t r = 0; r < new_screen_rows; r++) {
-        int UNUSED idx = (grid->view + r) & (new_rows - 1);
-        xassert(new_grid[idx] != NULL);
-    }
-
     /* Free old grid (rows already free:d) */
     free(grid->rows);
 
@@ -940,6 +1020,17 @@ grid_resize_and_reflow(
     grid->num_rows = new_rows;
     grid->num_cols = new_cols;
 
+    /*
+     * Set new viewport, making sure it’s not too far down.
+     *
+     * This is done by using scrollback-start relative cooardinates,
+     * and bounding the new viewport to (grid_rows - screen_rows).
+     */
+    int sb_view = grid_row_abs_to_sb(
+        grid, new_screen_rows, view_follows ? grid->offset : viewport.row);
+    grid->view = grid_row_sb_to_abs(
+        grid, new_screen_rows, min(sb_view, new_rows - new_screen_rows));
+
     /* Convert absolute coordinates to screen relative */
     cursor.row -= grid->offset;
     while (cursor.row < 0)
