@@ -17,17 +17,57 @@
 #include "wayland.h"
 #include "xmalloc.h"
 
+static void
+ime_reset_pending_preedit(struct seat *seat)
+{
+    free(seat->ime.preedit.pending.text);
+    seat->ime.preedit.pending.text = NULL;
+}
+
+static void
+ime_reset_pending_commit(struct seat *seat)
+{
+    free(seat->ime.commit.pending.text);
+    seat->ime.commit.pending.text = NULL;
+}
+
+void
+ime_reset_pending(struct seat *seat)
+{
+    ime_reset_pending_preedit(seat);
+    ime_reset_pending_commit(seat);
+}
+
+void
+ime_reset_preedit(struct seat *seat)
+{
+    if (seat->ime.preedit.cells == NULL)
+        return;
+
+    free(seat->ime.preedit.text);
+    free(seat->ime.preedit.cells);
+    seat->ime.preedit.text = NULL;
+    seat->ime.preedit.cells = NULL;
+    seat->ime.preedit.count = 0;
+}
+
 static void
 enter(void *data, struct zwp_text_input_v3 *zwp_text_input_v3,
       struct wl_surface *surface)
 {
     struct seat *seat = data;
+    struct wl_window *win = wl_surface_get_user_data(surface);
+    struct terminal *term = win->term;
 
-    LOG_DBG("enter: seat=%s", seat->name);
+    LOG_DBG("enter: seat=%s, term=%p", seat->name, (const void *)term);
+
+    if (seat->kbd_focus != term) {
+        LOG_WARN("compositor sent ime::enter() event before the "
+                 "corresponding keyboard_enter() event");
+    }
 
     /* The main grid is the *only* input-receiving surface we have */
-    xassert(seat->kbd_focus != NULL);
-    seat->ime.focused = true;
+    seat->ime_focus = term;
     ime_enable(seat);
 }
 
@@ -39,7 +79,7 @@ leave(void *data, struct zwp_text_input_v3 *zwp_text_input_v3,
     LOG_DBG("leave: seat=%s", seat->name);
 
     ime_disable(seat);
-    seat->ime.focused = false;
+    seat->ime_focus = NULL;
 }
 
 static void
@@ -104,7 +144,7 @@ done(void *data, struct zwp_text_input_v3 *zwp_text_input_v3,
 
     LOG_DBG("done: serial=%u", serial);
     struct seat *seat = data;
-    struct terminal *term = seat->kbd_focus;
+    struct terminal *term = seat->ime_focus;
 
     if (seat->ime.serial != serial) {
         LOG_DBG("IME serial mismatch: expected=0x%08x, got 0x%08x",
@@ -324,50 +364,16 @@ done(void *data, struct zwp_text_input_v3 *zwp_text_input_v3,
     }
 }
 
-void
-ime_reset_pending_preedit(struct seat *seat)
-{
-    free(seat->ime.preedit.pending.text);
-    seat->ime.preedit.pending.text = NULL;
-}
-
-void
-ime_reset_pending_commit(struct seat *seat)
-{
-    free(seat->ime.commit.pending.text);
-    seat->ime.commit.pending.text = NULL;
-}
-
-void
-ime_reset_pending(struct seat *seat)
-{
-    ime_reset_pending_preedit(seat);
-    ime_reset_pending_commit(seat);
-}
-
-void
-ime_reset_preedit(struct seat *seat)
-{
-    if (seat->ime.preedit.cells == NULL)
-        return;
-
-    free(seat->ime.preedit.text);
-    free(seat->ime.preedit.cells);
-    seat->ime.preedit.text = NULL;
-    seat->ime.preedit.cells = NULL;
-    seat->ime.preedit.count = 0;
-}
-
 static void
 ime_send_cursor_rect(struct seat *seat)
 {
     if (unlikely(seat->wayl->text_input_manager == NULL))
         return;
 
-    if (!seat->ime.focused)
+    if (seat->ime_focus == NULL)
         return;
 
-    struct terminal *term = seat->kbd_focus;
+    struct terminal *term = seat->ime_focus;
 
     if (!term->ime_enabled)
         return;
@@ -399,10 +405,10 @@ ime_enable(struct seat *seat)
     if (unlikely(seat->wayl->text_input_manager == NULL))
         return;
 
-    if (!seat->ime.focused)
+    if (seat->ime_focus == NULL)
         return;
 
-    struct terminal *term = seat->kbd_focus;
+    struct terminal *term = seat->ime_focus;
     if (term == NULL)
         return;
 
@@ -437,7 +443,7 @@ ime_disable(struct seat *seat)
     if (unlikely(seat->wayl->text_input_manager == NULL))
         return;
 
-    if (!seat->ime.focused)
+    if (seat->ime_focus == NULL)
         return;
 
     ime_reset_pending(seat);
@@ -451,7 +457,7 @@ ime_disable(struct seat *seat)
 void
 ime_update_cursor_rect(struct seat *seat)
 {
-    struct terminal *term = seat->kbd_focus;
+    struct terminal *term = seat->ime_focus;
 
     /* Set in render_ime_preedit() */
     if (seat->ime.preedit.cells != NULL)
