@@ -23,8 +23,9 @@
 #define LOG_MODULE "input"
 #define LOG_ENABLE_DBG 0
 #include "log.h"
-#include "config.h"
 #include "commands.h"
+#include "config.h"
+#include "grid.h"
 #include "keymap.h"
 #include "kitty-keymap.h"
 #include "macros.h"
@@ -35,6 +36,7 @@
 #include "spawn.h"
 #include "terminal.h"
 #include "tokenize.h"
+#include "unicode-mode.h"
 #include "url-mode.h"
 #include "util.h"
 #include "vt.h"
@@ -81,9 +83,10 @@ pipe_closed:
 
 static bool
 execute_binding(struct seat *seat, struct terminal *term,
-                enum bind_action_normal action, char *const *pipe_argv,
-                uint32_t serial)
+                const struct key_binding *binding, uint32_t serial)
 {
+    const enum bind_action_normal action = binding->action;
+
     switch (action) {
     case BIND_ACTION_NONE:
         return true;
@@ -133,6 +136,20 @@ execute_binding(struct seat *seat, struct terminal *term,
         }
         break;
 
+    case BIND_ACTION_SCROLLBACK_HOME:
+        if (term->grid == &term->normal) {
+            cmd_scrollback_up(term, term->grid->num_rows);
+            return true;
+        }
+        break;
+
+    case BIND_ACTION_SCROLLBACK_END:
+        if (term->grid == &term->normal) {
+            cmd_scrollback_down(term, term->grid->num_rows);
+            return true;
+        }
+        break;
+
     case BIND_ACTION_CLIPBOARD_COPY:
         selection_to_clipboard(seat, term, serial);
         return true;
@@ -193,7 +210,7 @@ execute_binding(struct seat *seat, struct terminal *term,
         /* FALLTHROUGH */
     case BIND_ACTION_PIPE_VIEW:
     case BIND_ACTION_PIPE_SELECTED: {
-        if (pipe_argv == NULL)
+        if (binding->aux->type != BINDING_AUX_PIPE)
             return true;
 
         struct pipe_context *ctx = NULL;
@@ -266,7 +283,8 @@ execute_binding(struct seat *seat, struct terminal *term,
             }
         }
 
-        if (!spawn(term->reaper, NULL, pipe_argv, pipe_fd[0], stdout_fd, stderr_fd))
+        if (!spawn(term->reaper, NULL, binding->aux->pipe.args,
+                   pipe_fd[0], stdout_fd, stderr_fd, NULL))
             goto pipe_err;
 
         /* Close read end */
@@ -299,12 +317,14 @@ execute_binding(struct seat *seat, struct terminal *term,
     }
 
     case BIND_ACTION_SHOW_URLS_COPY:
-    case BIND_ACTION_SHOW_URLS_LAUNCH: {
+    case BIND_ACTION_SHOW_URLS_LAUNCH:
+    case BIND_ACTION_SHOW_URLS_PERSISTENT: {
         xassert(!urls_mode_is_active(term));
 
-        enum url_action url_action = action == BIND_ACTION_SHOW_URLS_COPY
-            ? URL_ACTION_COPY
-            : URL_ACTION_LAUNCH;
+        enum url_action url_action =
+            action == BIND_ACTION_SHOW_URLS_COPY ? URL_ACTION_COPY :
+            action == BIND_ACTION_SHOW_URLS_LAUNCH ? URL_ACTION_LAUNCH :
+            URL_ACTION_PERSISTENT;
 
         urls_collect(term, url_action, &term->urls);
         urls_assign_key_combos(term->conf, &term->urls);
@@ -312,6 +332,95 @@ execute_binding(struct seat *seat, struct terminal *term,
         return true;
     }
 
+    case BIND_ACTION_TEXT_BINDING:
+        xassert(binding->aux->type == BINDING_AUX_TEXT);
+        term_to_slave(term, binding->aux->text.data, binding->aux->text.len);
+        return true;
+
+    case BIND_ACTION_PROMPT_PREV: {
+        if (term->grid != &term->normal)
+            return false;
+
+        struct grid *grid = term->grid;
+        const int sb_start =
+            grid_sb_start_ignore_uninitialized(grid, term->rows);
+
+        /* Check each row from current view-1 (that is, the first
+         * currently not visible row), up to, and including, the
+         * scrollback start */
+        for (int r_sb_rel =
+                 grid_row_abs_to_sb_precalc_sb_start(
+                     grid, sb_start, grid->view) - 1;
+             r_sb_rel >= 0; r_sb_rel--)
+        {
+            const int r_abs =
+                grid_row_sb_to_abs_precalc_sb_start(grid, sb_start, r_sb_rel);
+
+            const struct row *row = grid->rows[r_abs];
+            xassert(row != NULL);
+
+            if (!row->prompt_marker)
+                continue;
+
+            grid->view = r_abs;
+            term_damage_view(term);
+            render_refresh(term);
+            break;
+        }
+
+        return true;
+    }
+
+    case BIND_ACTION_PROMPT_NEXT: {
+        if (term->grid != &term->normal)
+            return false;
+
+        struct grid *grid = term->grid;
+        const int num_rows = grid->num_rows;
+
+        if (grid->view == grid->offset) {
+            /* Already at the bottom */
+            return true;
+        }
+
+        /* Check each row from view+1, to the bottom of the scrollback */
+        for (int r_abs = (grid->view + 1) & (num_rows - 1);
+             ;
+             r_abs = (r_abs + 1) & (num_rows - 1))
+        {
+            const struct row *row = grid->rows[r_abs];
+            xassert(row != NULL);
+
+            if (!row->prompt_marker) {
+                if (r_abs == grid->offset + term->rows - 1) {
+                    /* We’ve reached the bottom of the scrollback */
+                    break;
+                }
+                continue;
+            }
+
+            int sb_start = grid_sb_start_ignore_uninitialized(grid, term->rows);
+            int ofs_sb_rel =
+                grid_row_abs_to_sb_precalc_sb_start(grid, sb_start, grid->offset);
+            int new_view_sb_rel =
+                grid_row_abs_to_sb_precalc_sb_start(grid, sb_start, r_abs);
+
+            new_view_sb_rel = min(ofs_sb_rel, new_view_sb_rel);
+            grid->view = grid_row_sb_to_abs_precalc_sb_start(
+                grid, sb_start, new_view_sb_rel);
+
+            term_damage_view(term);
+            render_refresh(term);
+            break; 
+        }
+
+        return true;
+    }
+
+    case BIND_ACTION_UNICODE_INPUT:
+        unicode_mode_activate(seat);
+        return true;
+
     case BIND_ACTION_SELECT_BEGIN:
         selection_start(
             term, seat->mouse.col, seat->mouse.row, SELECTION_CHAR_WISE, false);
@@ -358,231 +467,6 @@ execute_binding(struct seat *seat, struct terminal *term,
     return false;
 }
 
-static xkb_mod_mask_t
-conf_modifiers_to_mask(const struct seat *seat,
-                       const struct config_key_modifiers *modifiers)
-{
-    xkb_mod_mask_t mods = 0;
-    if (seat->kbd.mod_shift != XKB_MOD_INVALID)
-        mods |= modifiers->shift << seat->kbd.mod_shift;
-    if (seat->kbd.mod_ctrl != XKB_MOD_INVALID)
-        mods |= modifiers->ctrl << seat->kbd.mod_ctrl;
-    if (seat->kbd.mod_alt != XKB_MOD_INVALID)
-        mods |= modifiers->alt << seat->kbd.mod_alt;
-    if (seat->kbd.mod_super != XKB_MOD_INVALID)
-        mods |= modifiers->super << seat->kbd.mod_super;
-    return mods;
-}
-
-static xkb_keycode_list_t
-key_codes_for_xkb_sym(struct xkb_keymap *keymap, xkb_keysym_t sym)
-{
-    xkb_keycode_list_t key_codes = tll_init();
-
-    /*
-     * Find all key codes that map to this symbol.
-     *
-     * This allows us to match bindings in other layouts
-     * too.
-     */
-    struct xkb_state *state = xkb_state_new(keymap);
-
-    for (xkb_keycode_t code = xkb_keymap_min_keycode(keymap);
-         code <= xkb_keymap_max_keycode(keymap);
-         code++)
-    {
-        if (xkb_state_key_get_one_sym(state, code) == sym)
-            tll_push_back(key_codes, code);
-    }
-
-    xkb_state_unref(state);
-    return key_codes;
-}
-
-static xkb_keysym_t
-maybe_repair_key_combo(const struct seat *seat,
-                       xkb_keysym_t sym, xkb_mod_mask_t mods)
-{
-    /*
-     * Detect combos containing a shifted symbol and the corresponding
-     * modifier, and replace the shifted symbol with its unshifted
-     * variant.
-     *
-     * For example, the combo is “Control+Shift+U”. In this case,
-     * Shift is the modifier used to “shift” ‘u’ to ‘U’, after which
-     * ‘Shift’ will have been “consumed”. Since we filter out consumed
-     * modifiers when matching key combos, this key combo will never
-     * trigger (we will never be able to match the ‘Shift’ modifier).
-     *
-     * There are two correct variants of the above key combo:
-     *  - “Control+U”           (upper case ‘U’)
-     *  - “Control+Shift+u”     (lower case ‘u’)
-     *
-     * What we do here is, for each key *code*, check if there are any
-     * (shifted) levels where it produces ‘sym’. If there are, check
-     * *which* sets of modifiers are needed to produce it, and compare
-     * with ‘mods’.
-     *
-     * If there is at least one common modifier, it means ‘sym’ is a
-     * “shifted” symbol, with the corresponding shifting modifier
-     * explicitly included in the key combo. I.e. the key combo will
-     * never trigger.
-     *
-     * We then proceed and “repair” the key combo by replacing ‘sym’
-     * with the corresponding unshifted symbol.
-     *
-     * To reduce the noise, we ignore all key codes where the shifted
-     * symbol is the same as the unshifted symbol.
-     */
-
-    for (xkb_keycode_t code = xkb_keymap_min_keycode(seat->kbd.xkb_keymap);
-         code <= xkb_keymap_max_keycode(seat->kbd.xkb_keymap);
-         code++)
-    {
-        xkb_layout_index_t layout_idx =
-            xkb_state_key_get_layout(seat->kbd.xkb_state, code);
-
-        /* Get all unshifted symbols for this key */
-        const xkb_keysym_t *base_syms = NULL;
-        size_t base_count = xkb_keymap_key_get_syms_by_level(
-            seat->kbd.xkb_keymap, code, layout_idx, 0, &base_syms);
-
-        if (base_count == 0 || sym == base_syms[0]) {
-            /* No unshifted symbols, or unshifted symbol is same as ‘sym’ */
-            continue;
-        }
-
-        /* Name of the unshifted symbol, for logging */
-        char base_name[100];
-        xkb_keysym_get_name(base_syms[0], base_name, sizeof(base_name));
-
-        /* Iterate all shift levels */
-        for (xkb_level_index_t level_idx = 1;
-             level_idx < xkb_keymap_num_levels_for_key(
-                 seat->kbd.xkb_keymap, code, layout_idx);
-             level_idx++) {
-
-            /* Get all symbols for current shift level */
-            const xkb_keysym_t *shifted_syms = NULL;
-            size_t shifted_count = xkb_keymap_key_get_syms_by_level(
-                seat->kbd.xkb_keymap, code,
-                layout_idx, level_idx, &shifted_syms);
-
-            for (size_t i = 0; i < shifted_count; i++) {
-                if (shifted_syms[i] != sym)
-                    continue;
-
-                /* Get modifier sets that produces the current shift level */
-                xkb_mod_mask_t mod_masks[16];
-                size_t mod_mask_count = xkb_keymap_key_get_mods_for_level(
-                    seat->kbd.xkb_keymap, code, layout_idx, level_idx,
-                    mod_masks, ALEN(mod_masks));
-
-                /* Check if key combo’s modifier set intersects */
-                for (size_t j = 0; j < mod_mask_count; j++) {
-                    if ((mod_masks[j] & mods) != mod_masks[j])
-                        continue;
-
-                    char combo[64] = {0};
-
-                    for (int k = 0; k < sizeof(xkb_mod_mask_t) * 8; k++) {
-                        if (!(mods & (1u << k)))
-                            continue;
-
-                        const char *mod_name = xkb_keymap_mod_get_name(
-                            seat->kbd.xkb_keymap, k);
-                        strcat(combo, mod_name);
-                        strcat(combo, "+");
-                    }
-
-                    size_t len = strlen(combo);
-                    xkb_keysym_get_name(
-                        sym, &combo[len], sizeof(combo) - len);
-
-                    LOG_WARN(
-                        "%s: combo with both explicit modifier and shifted symbol "
-                        "(level=%d, mod-mask=0x%08x), "
-                        "replacing with %s",
-                        combo, level_idx, mod_masks[j], base_name);
-
-                    /* Replace with unshifted symbol */
-                    return base_syms[0];
-                }
-            }
-        }
-    }
-
-    return sym;
-}
-
-static void
-convert_key_binding(const struct seat *seat,
-                    const struct config_key_binding *conf_binding,
-                    key_binding_list_t *bindings)
-{
-    xkb_mod_mask_t mods = conf_modifiers_to_mask(seat, &conf_binding->modifiers);
-    xkb_keysym_t sym = maybe_repair_key_combo(seat, conf_binding->k.sym, mods);
-
-    struct key_binding binding = {
-        .mods = mods,
-        .sym = sym,
-        .key_codes = key_codes_for_xkb_sym(seat->kbd.xkb_keymap, sym),
-        .action = conf_binding->action,
-        .pipe_argv = conf_binding->pipe.argv.args,
-    };
-    tll_push_back(*bindings, binding);
-}
-
-static void
-convert_key_bindings(const struct config *conf, struct seat *seat)
-{
-    for (size_t i = 0; i < conf->bindings.key.count; i++) {
-        const struct config_key_binding *binding = &conf->bindings.key.arr[i];
-        convert_key_binding(seat, binding, &seat->kbd.bindings.key);
-    }
-}
-
-static void
-convert_search_bindings(const struct config *conf, struct seat *seat)
-{
-    for (size_t i = 0; i < conf->bindings.search.count; i++) {
-        const struct config_key_binding *binding = &conf->bindings.search.arr[i];
-        convert_key_binding(seat, binding, &seat->kbd.bindings.search);
-    }
-}
-
-static void
-convert_url_bindings(const struct config *conf, struct seat *seat)
-{
-    for (size_t i = 0; i < conf->bindings.url.count; i++) {
-        const struct config_key_binding *binding = &conf->bindings.url.arr[i];
-        convert_key_binding(seat, binding, &seat->kbd.bindings.url);
-    }
-}
-
-static void
-convert_mouse_binding(struct seat *seat,
-                      const struct config_key_binding *conf_binding)
-{
-    struct mouse_binding binding = {
-        .action = conf_binding->action,
-        .mods = conf_modifiers_to_mask(seat, &conf_binding->modifiers),
-        .button = conf_binding->m.button,
-        .count = conf_binding->m.count,
-        .pipe_argv = conf_binding->pipe.argv.args,
-    };
-    tll_push_back(seat->mouse.bindings, binding);
-}
-
-static void
-convert_mouse_bindings(const struct config *conf, struct seat *seat)
-{
-    for (size_t i = 0; i < conf->bindings.mouse.count; i++) {
-        const struct config_key_binding *binding = &conf->bindings.mouse.arr[i];
-        convert_mouse_binding(seat, binding);
-    }
-}
-
 static void
 keyboard_keymap(void *data, struct wl_keyboard *wl_keyboard,
                 uint32_t format, int32_t fd, uint32_t size)
@@ -618,15 +502,7 @@ keyboard_keymap(void *data, struct wl_keyboard *wl_keyboard,
         seat->kbd.xkb = NULL;
     }
 
-    tll_foreach(seat->kbd.bindings.key, it)
-        tll_free(it->item.key_codes);
-    tll_free(seat->kbd.bindings.key);
-
-    tll_foreach(seat->kbd.bindings.search, it)
-        tll_free(it->item.key_codes);
-    tll_free(seat->kbd.bindings.search);
-
-    tll_free(seat->mouse.bindings);
+    key_binding_unload_keymap(wayl->key_binding_manager, seat);
 
     /* Verify keymap is in a format we understand */
     switch ((enum wl_keyboard_keymap_format)format) {
@@ -700,18 +576,10 @@ keyboard_keymap(void *data, struct wl_keyboard *wl_keyboard,
         seat->kbd.key_arrow_down = xkb_keymap_key_by_name(seat->kbd.xkb_keymap, "DOWN");
     }
 
-    /* Set selection-override modmask from configured mods and seat's mod indices */
-    const struct config_key_modifiers* override_mods =
-        &wayl->conf->mouse.selection_override_modifiers;
-    seat->kbd.selection_override_modmask = conf_modifiers_to_mask(seat, override_mods);
-
     munmap(map_str, size);
     close(fd);
 
-    convert_key_bindings(wayl->conf, seat);
-    convert_search_bindings(wayl->conf, seat);
-    convert_url_bindings(wayl->conf, seat);
-    convert_mouse_bindings(wayl->conf, seat);
+    key_binding_load_keymap(wayl->key_binding_manager, seat);
 }
 
 static void
@@ -719,6 +587,7 @@ keyboard_enter(void *data, struct wl_keyboard *wl_keyboard, uint32_t serial,
                struct wl_surface *surface, struct wl_array *keys)
 {
     xassert(surface != NULL);
+    xassert(serial != 0);
 
     struct seat *seat = data;
     struct wl_window *win = wl_surface_get_user_data(surface);
@@ -727,9 +596,6 @@ keyboard_enter(void *data, struct wl_keyboard *wl_keyboard, uint32_t serial,
     LOG_DBG("%s: keyboard_enter: keyboard=%p, serial=%u, surface=%p",
             seat->name, (void *)wl_keyboard, serial, (void *)surface);
 
-    if (seat->kbd.xkb == NULL)
-        return;
-
     term_kbd_focus_in(term);
     seat->kbd_focus = term;
     seat->kbd.serial = serial;
@@ -789,9 +655,6 @@ keyboard_leave(void *data, struct wl_keyboard *wl_keyboard, uint32_t serial,
     LOG_DBG("keyboard_leave: keyboard=%p, serial=%u, surface=%p",
             (void *)wl_keyboard, serial, (void *)surface);
 
-    if (seat->kbd.xkb == NULL)
-        return;
-
     xassert(
         seat->kbd_focus == NULL ||
         surface == NULL ||  /* Seen on Sway 1.2 */
@@ -992,17 +855,32 @@ get_current_modifiers(const struct seat *seat,
                       xkb_mod_mask_t *effective,
                       xkb_mod_mask_t *consumed, uint32_t key)
 {
-    if (effective != NULL) {
-        *effective = xkb_state_serialize_mods(
-            seat->kbd.xkb_state, XKB_STATE_MODS_EFFECTIVE);
+    if (unlikely(seat->kbd.xkb_state == NULL)) {
+        if (effective != NULL)
+            *effective = 0;
+        if (consumed != NULL)
+            *consumed = 0;
     }
 
-    if (consumed != NULL) {
-        *consumed = xkb_state_key_get_consumed_mods2(
-            seat->kbd.xkb_state, key, XKB_CONSUMED_MODE_XKB);
+    else {
+        if (effective != NULL) {
+            *effective = xkb_state_serialize_mods(
+                seat->kbd.xkb_state, XKB_STATE_MODS_EFFECTIVE);
+        }
+
+        if (consumed != NULL) {
+            *consumed = xkb_state_key_get_consumed_mods2(
+                seat->kbd.xkb_state, key, XKB_CONSUMED_MODE_XKB);
+        }
     }
 }
 
+static xkb_mod_mask_t
+get_locked_modifiers(const struct seat *seat)
+{
+    return xkb_state_serialize_mods(seat->kbd.xkb_state, XKB_STATE_MODS_LOCKED);
+}
+
 struct kbd_ctx {
     xkb_layout_index_t layout;
     xkb_keycode_t key;
@@ -1045,13 +923,7 @@ legacy_kbd_protocol(struct seat *seat, struct terminal *term,
     const size_t count = ctx->utf8.count;
     const uint8_t *const utf8 = ctx->utf8.buf;
 
-    const struct key_data *keymap;
-    if (sym == XKB_KEY_Escape && keymap_mods == MOD_NONE && term->modify_escape_key) {
-        static const struct key_data esc = {.seq = "\033[27;1;27~"};
-        keymap = &esc;
-    } else
-        keymap = keymap_lookup(term, sym, keymap_mods);
-
+    const struct key_data *keymap = keymap_lookup(term, sym, keymap_mods);
     if (keymap != NULL) {
         term_to_slave(term, keymap->seq, strlen(keymap->seq));
         return true;
@@ -1069,7 +941,56 @@ legacy_kbd_protocol(struct seat *seat, struct terminal *term,
     bool ctrl_is_in_effect = (keymap_mods & MOD_CTRL) != 0;
     bool ctrl_seq = is_control_key(sym) || (count == 1 && IS_CTRL(utf8[0]));
 
-    if (keymap_mods != MOD_NONE && (term->modify_other_keys_2 ||
+    bool modify_other_keys2_in_effect = false;
+
+    if (term->modify_other_keys_2) {
+        /*
+         * Try to mimic XTerm’s behavior, when holding shift:
+         *
+         *   - if other modifiers are pressed (e.g. Alt), emit a CSI escape
+         *   - upper-case symbols A-Z are encoded as an CSI escape
+         *   - other upper-case symbols (e.g ‘Ö’) or emitted as is
+         *   - non-upper cased symbols are _mostly_ emitted as is (foot
+         *     always emits as is)
+         *
+         * Examples (assuming Swedish layout):
+         *   - Shift-a (‘A’) emits a CSI
+         *   - Shift-, (‘;’) emits ‘;’
+         *   - Shift-Alt-, (Alt-;) emits a CSI
+         *   - Shift-ö (‘Ö’) emits ‘Ö’
+         */
+
+        /* Any modifiers, besides shift active? */
+        const xkb_mod_mask_t shift_mask = 1 << seat->kbd.mod_shift;
+        if ((ctx->mods & ~shift_mask & seat->kbd.bind_significant) != 0)
+            modify_other_keys2_in_effect = true;
+
+        else {
+            const xkb_layout_index_t layout_idx = xkb_state_key_get_layout(
+                seat->kbd.xkb_state, ctx->key);
+
+            /*
+             * Get pressed key’s base symbol.
+             *   - for ‘A’ (shift-a), that’s ‘a’
+             *   - for ‘;’ (shift-,), that’s ‘,’
+             */
+            const xkb_keysym_t *base_syms = NULL;
+            size_t base_count = xkb_keymap_key_get_syms_by_level(
+                seat->kbd.xkb_keymap, ctx->key, layout_idx, 0, &base_syms);
+
+            /* Check if base symbol(s) is a-z. If so, emit CSI */
+            const xkb_keysym_t lower_cased_sym = xkb_keysym_to_lower(ctx->sym);
+            for (size_t i = 0; i < base_count; i++) {
+                const xkb_keysym_t s = base_syms[i];
+                if (lower_cased_sym == s && s >= XKB_KEY_a && s <= XKB_KEY_z) {
+                    modify_other_keys2_in_effect = true;
+                    break;
+                }
+            }
+        }
+    }
+
+    if (keymap_mods != MOD_NONE && (modify_other_keys2_in_effect ||
                                     (ctrl_is_in_effect && !ctrl_seq)))
     {
         static const int mod_param_map[32] = {
@@ -1323,7 +1244,7 @@ emit_escapes:
             ? ctx->level0_syms.syms[0]
             : sym;
 
-        if (composed && is_text)
+        if (composed)
             key = utf32;
         else {
             key = xkb_keysym_to_utf32(sym_to_use);
@@ -1414,10 +1335,24 @@ emit_escapes:
     return term_to_slave(term, buf, sizeof(buf) - left);
 }
 
+/* Copied from libxkbcommon (internal function) */
+static bool
+keysym_is_modifier(xkb_keysym_t keysym)
+{
+    return
+        (keysym >= XKB_KEY_Shift_L && keysym <= XKB_KEY_Hyper_R) ||
+        /* libX11 only goes up to XKB_KEY_ISO_Level5_Lock. */
+        (keysym >= XKB_KEY_ISO_Lock && keysym <= XKB_KEY_ISO_Last_Group_Lock) ||
+        keysym == XKB_KEY_Mode_switch ||
+        keysym == XKB_KEY_Num_Lock;
+}
+
 static void
 key_press_release(struct seat *seat, struct terminal *term, uint32_t serial,
                   uint32_t key, uint32_t state)
 {
+    xassert(serial != 0);
+
     seat->kbd.serial = serial;
     if (seat->kbd.xkb == NULL ||
         seat->kbd.xkb_keymap == NULL ||
@@ -1438,17 +1373,7 @@ key_press_release(struct seat *seat, struct terminal *term, uint32_t serial,
 
     xkb_keysym_t sym = xkb_state_key_get_one_sym(seat->kbd.xkb_state, key);
 
-    if (pressed && term->conf->mouse.hide_when_typing &&
-
-        /* TODO: better way to detect modifiers */
-        sym != XKB_KEY_Shift_L && sym != XKB_KEY_Shift_R &&
-        sym != XKB_KEY_Control_L && sym != XKB_KEY_Control_R &&
-        sym != XKB_KEY_Alt_L && sym != XKB_KEY_Alt_R &&
-        sym != XKB_KEY_ISO_Level3_Shift &&
-        sym != XKB_KEY_Super_L && sym != XKB_KEY_Super_R &&
-        sym != XKB_KEY_Meta_L && sym != XKB_KEY_Meta_R &&
-        sym != XKB_KEY_Menu)
-    {
+    if (pressed && term->conf->mouse.hide_when_typing && !keysym_is_modifier(sym)) {
         seat->pointer.hidden = true;
         term_xcursor_update_for_seat(term, seat);
     }
@@ -1467,8 +1392,11 @@ key_press_release(struct seat *seat, struct terminal *term, uint32_t serial,
     xkb_mod_mask_t mods, consumed;
     get_current_modifiers(seat, &mods, &consumed, key);
 
-    xkb_mod_mask_t bind_mods = mods & seat->kbd.bind_significant;
-    xkb_mod_mask_t bind_consumed = consumed & seat->kbd.bind_significant;
+    const xkb_mod_mask_t locked = get_locked_modifiers(seat);
+    const xkb_mod_mask_t bind_mods
+        = mods & seat->kbd.bind_significant & ~locked;
+    const xkb_mod_mask_t bind_consumed =
+        consumed & seat->kbd.bind_significant & ~locked;
 
     xkb_layout_index_t layout_idx =
         xkb_state_key_get_layout(seat->kbd.xkb_state, key);
@@ -1477,13 +1405,22 @@ key_press_release(struct seat *seat, struct terminal *term, uint32_t serial,
     size_t raw_count = xkb_keymap_key_get_syms_by_level(
         seat->kbd.xkb_keymap, key, layout_idx, 0, &raw_syms);
 
+    const struct key_binding_set *bindings = key_binding_for(
+        seat->wayl->key_binding_manager, term, seat);
+    xassert(bindings != NULL);
+
     if (pressed) {
-        if (term->is_searching) {
+        if (seat->unicode_mode.active) {
+            unicode_mode_input(seat, term, sym);
+            return;
+        }
+
+        else if (term->is_searching) {
             if (should_repeat)
                 start_repeater(seat, key);
 
             search_input(
-                seat, term, key, sym, bind_mods, bind_consumed,
+                seat, term, bindings, key, sym, mods, consumed, locked,
                 raw_syms, raw_count, serial);
             return;
         }
@@ -1493,7 +1430,7 @@ key_press_release(struct seat *seat, struct terminal *term, uint32_t serial,
                 start_repeater(seat, key);
 
             urls_input(
-                seat, term, key, sym, bind_mods, bind_consumed,
+                seat, term, bindings, key, sym, mods, consumed, locked,
                 raw_syms, raw_count, serial);
             return;
         }
@@ -1521,34 +1458,33 @@ key_press_release(struct seat *seat, struct terminal *term, uint32_t serial,
      * User configurable bindings
      */
     if (pressed) {
-        tll_foreach(seat->kbd.bindings.key, it) {
+        tll_foreach(bindings->key, it) {
             const struct key_binding *bind = &it->item;
 
             /* Match translated symbol */
-            if (bind->sym == sym &&
+            if (bind->k.sym == sym &&
                 bind->mods == (bind_mods & ~bind_consumed) &&
-                execute_binding(
-                    seat, term, bind->action, bind->pipe_argv, serial))
+                execute_binding(seat, term, bind, serial))
             {
                 goto maybe_repeat;
             }
 
-            if (bind->mods != bind_mods)
+            if (bind->mods != bind_mods || bind_mods != (mods & ~locked))
                 continue;
 
             /* Match untranslated symbols */
             for (size_t i = 0; i < raw_count; i++) {
-                if (bind->sym == raw_syms[i] && execute_binding(
-                        seat, term, bind->action, bind->pipe_argv, serial))
+                if (bind->k.sym == raw_syms[i] &&
+                    execute_binding(seat, term, bind, serial))
                 {
                     goto maybe_repeat;
                 }
             }
 
             /* Match raw key code */
-            tll_foreach(bind->key_codes, code) {
-                if (code->item == key && execute_binding(
-                        seat, term, bind->action, bind->pipe_argv, serial))
+            tll_foreach(bind->k.key_codes, code) {
+                if (code->item == key &&
+                    execute_binding(seat, term, bind, serial))
                 {
                     goto maybe_repeat;
                 }
@@ -1773,10 +1709,9 @@ wl_pointer_enter(void *data, struct wl_pointer *wl_pointer,
                  uint32_t serial, struct wl_surface *surface,
                  wl_fixed_t surface_x, wl_fixed_t surface_y)
 {
-    xassert(surface != NULL);
-
-    if (surface == NULL)  {
+    if (unlikely(surface == NULL))  {
         /* Seen on mutter-3.38 */
+        LOG_WARN("compositor sent pointer_enter event with a NULL surface");
         return;
     }
 
@@ -1826,10 +1761,6 @@ wl_pointer_enter(void *data, struct wl_pointer *wl_pointer,
         break;
     }
 
-    case TERM_SURF_SEARCH:
-    case TERM_SURF_SCROLLBACK_INDICATOR:
-    case TERM_SURF_RENDER_TIMER:
-    case TERM_SURF_JUMP_LABEL:
     case TERM_SURF_TITLE:
     case TERM_SURF_BORDER_LEFT:
     case TERM_SURF_BORDER_RIGHT:
@@ -1915,10 +1846,6 @@ wl_pointer_leave(void *data, struct wl_pointer *wl_pointer,
             break;
 
         case TERM_SURF_NONE:
-        case TERM_SURF_SEARCH:
-        case TERM_SURF_SCROLLBACK_INDICATOR:
-        case TERM_SURF_RENDER_TIMER:
-        case TERM_SURF_JUMP_LABEL:
         case TERM_SURF_TITLE:
         case TERM_SURF_BORDER_LEFT:
         case TERM_SURF_BORDER_RIGHT:
@@ -1937,6 +1864,16 @@ wl_pointer_motion(void *data, struct wl_pointer *wl_pointer,
     struct seat *seat = data;
     struct wayland *wayl = seat->wayl;
     struct terminal *term = seat->mouse_focus;
+
+    if (unlikely(term == NULL)) {
+        /* Typically happens when the compositor sent a pointer enter
+         * event with a NULL surface - see wl_pointer_enter().
+         *
+         * In this case, we never set seat->mouse_focus (since we
+         * can’t map the enter event to a specific window). */
+        return;
+    }
+
     struct wl_window *win = term->window;
 
     LOG_DBG("pointer_motion: pointer=%p, x=%d, y=%d", (void *)wl_pointer,
@@ -1966,10 +1903,6 @@ wl_pointer_motion(void *data, struct wl_pointer *wl_pointer,
 
     switch (surf_kind) {
     case TERM_SURF_NONE:
-    case TERM_SURF_SEARCH:
-    case TERM_SURF_SCROLLBACK_INDICATOR:
-    case TERM_SURF_RENDER_TIMER:
-    case TERM_SURF_JUMP_LABEL:
     case TERM_SURF_BUTTON_MINIMIZE:
     case TERM_SURF_BUTTON_MAXIMIZE:
     case TERM_SURF_BUTTON_CLOSE:
@@ -2092,8 +2025,9 @@ wl_pointer_motion(void *data, struct wl_pointer *wl_pointer,
                     auto_scroll_direction, selection_col);
             }
 
-            if (term->selection.ongoing && (cursor_is_on_new_cell ||
-                                            term->selection.end.row < 0))
+            if (term->selection.ongoing && (
+                    cursor_is_on_new_cell ||
+                    term->selection.coords.end.row < 0))
             {
                 selection_update(term, selection_col, selection_row);
             }
@@ -2148,6 +2082,8 @@ wl_pointer_button(void *data, struct wl_pointer *wl_pointer,
     LOG_DBG("BUTTON: pointer=%p, serial=%u, button=%x, state=%u",
             (void *)wl_pointer, serial, button, state);
 
+    xassert(serial != 0);
+
     struct seat *seat = data;
     struct wayland *wayl = seat->wayl;
     struct terminal *term = seat->mouse_focus;
@@ -2384,12 +2320,6 @@ wl_pointer_button(void *data, struct wl_pointer *wl_pointer,
             term_shutdown(term);
         break;
 
-    case TERM_SURF_SEARCH:
-    case TERM_SURF_SCROLLBACK_INDICATOR:
-    case TERM_SURF_RENDER_TIMER:
-    case TERM_SURF_JUMP_LABEL:
-        break;
-
     case TERM_SURF_GRID: {
         search_cancel(term);
         urls_reset(term);
@@ -2404,19 +2334,24 @@ wl_pointer_button(void *data, struct wl_pointer *wl_pointer,
                 if (seat->wl_keyboard != NULL && seat->kbd.xkb_state != NULL) {
                     /* Seat has keyboard - use mouse bindings *with* modifiers */
 
+                    const struct key_binding_set *bindings = key_binding_for(
+                        wayl->key_binding_manager, term, seat);
+                    xassert(bindings != NULL);
+
                     xkb_mod_mask_t mods;
                     get_current_modifiers(seat, &mods, NULL, 0);
                     mods &= seat->kbd.bind_significant;
 
-                    /* Ignore selection override modifiers when matching modifiers */
-                    mods &= ~seat->kbd.selection_override_modmask;
+                    /* Ignore selection override modifiers when
+                     * matching modifiers */
+                    mods &= ~bindings->selection_overrides;
 
-                    const struct mouse_binding *match = NULL;
+                    const struct key_binding *match = NULL;
 
-                    tll_foreach(seat->mouse.bindings, it) {
-                        const struct mouse_binding *binding = &it->item;
+                    tll_foreach(bindings->mouse, it) {
+                        const struct key_binding *binding = &it->item;
 
-                        if (binding->button != button) {
+                        if (binding->m.button != button) {
                             /* Wrong button */
                             continue;
                         }
@@ -2426,25 +2361,23 @@ wl_pointer_button(void *data, struct wl_pointer *wl_pointer,
                             continue;
                         }
 
-                        if  (binding->count > seat->mouse.count) {
+                        if  (binding->m.count > seat->mouse.count) {
                             /* Not correct click count */
                             continue;
                         }
 
-                        if (match == NULL || binding->count > match->count)
+                        if (match == NULL || binding->m.count > match->m.count)
                             match = binding;
                     }
 
-                    if (match != NULL) {
-                        consumed = execute_binding(
-                            seat, term, match->action, match->pipe_argv, serial);
-                    }
+                    if (match != NULL)
+                        consumed = execute_binding(seat, term, match, serial);
                 }
 
                 else {
                     /* Seat does NOT have a keyboard - use mouse bindings *without* modifiers */
                     const struct config_key_binding *match = NULL;
-                    const struct config *conf = seat->wayl->conf;
+                    const struct config *conf = term->conf;
 
                     for (size_t i = 0; i < conf->bindings.mouse.count; i++) {
                         const struct config_key_binding *binding =
@@ -2471,9 +2404,11 @@ wl_pointer_button(void *data, struct wl_pointer *wl_pointer,
                     }
 
                     if (match != NULL) {
-                        consumed = execute_binding(
-                            seat, term, match->action, match->pipe.argv.args,
-                            serial);
+                        struct key_binding bind = {
+                            .action = match->action,
+                            .aux = &match->aux,
+                        };
+                        consumed = execute_binding(seat, term, &bind, serial);
                     }
                 }
             }
