@@ -21,8 +21,9 @@
 #include "log.h"
 
 #include "config.h"
-#include "foot-features.h"
 #include "fdm.h"
+#include "foot-features.h"
+#include "key-binding.h"
 #include "macros.h"
 #include "reaper.h"
 #include "render.h"
@@ -454,12 +455,12 @@ main(int argc, char *const *argv)
             const char *const fallback_locale = fallback_locales[i];
 
             if (setlocale(LC_CTYPE, fallback_locale) != NULL) {
-                LOG_WARN("locale '%s' is not UTF-8, using '%s' instead",
+                LOG_WARN("'%s' is not a UTF-8 locale, using '%s' instead",
                          locale, fallback_locale);
 
                 user_notification_add_fmt(
                     &user_notifications, USER_NOTIFICATION_WARNING,
-                    "locale '%s' is not UTF-8, using '%s' instead",
+                    "'%s' is not a UTF-8 locale, using '%s' instead",
                     locale, fallback_locale);
 
                 bad_locale = false;
@@ -468,13 +469,13 @@ main(int argc, char *const *argv)
         }
 
         if (bad_locale) {
-            LOG_ERR("locale '%s' is not UTF-8, "
-                    "and failed to enable a fallback locale", locale);
+            LOG_ERR(
+                "'%s' is not a UTF-8 locale, and failed to find a fallback",
+                locale);
 
             user_notification_add_fmt(
                 &user_notifications, USER_NOTIFICATION_ERROR,
-                "locale '%s' is not UTF-8, "
-                "and failed to enable a fallback locale",
+                "'%s' is not a UTF-8 locale, and failed to find a fallback",
                 locale);
         }
     }
@@ -485,12 +486,12 @@ main(int argc, char *const *argv)
 
     tll_free(overrides);
     if (!conf_successful) {
-        config_free(conf);
+        config_free(&conf);
         return ret;
     }
 
     if (check_config) {
-        config_free(conf);
+        config_free(&conf);
         return EXIT_SUCCESS;
     }
 
@@ -569,6 +570,7 @@ main(int argc, char *const *argv)
 
     struct fdm *fdm = NULL;
     struct reaper *reaper = NULL;
+    struct key_binding_manager *key_binding_manager = NULL;
     struct wayland *wayl = NULL;
     struct renderer *renderer = NULL;
     struct terminal *term = NULL;
@@ -600,15 +602,21 @@ main(int argc, char *const *argv)
     if ((reaper = reaper_init(fdm)) == NULL)
         goto out;
 
-    if ((wayl = wayl_init(&conf, fdm)) == NULL)
+    if ((key_binding_manager = key_binding_manager_new()) == NULL)
         goto out;
 
+    if ((wayl = wayl_init(
+             fdm, key_binding_manager, conf.presentation_timings)) == NULL)
+    {
+        goto out;
+    }
+
     if ((renderer = render_init(fdm, wayl)) == NULL)
         goto out;
 
     if (!as_server && (term = term_init(
                            &conf, fdm, reaper, wayl, "foot", cwd, token,
-                           argc, argv,
+                           argc, argv, NULL,
                            &term_shutdown_cb, &shutdown_ctx)) == NULL) {
         goto out;
     }
@@ -625,7 +633,8 @@ main(int argc, char *const *argv)
         goto out;
     }
 
-    const struct sigaction sig_ign = {.sa_handler = SIG_IGN};
+    struct sigaction sig_ign = {.sa_handler = SIG_IGN};
+    sigemptyset(&sig_ign.sa_mask);
     if (sigaction(SIGHUP, &sig_ign, NULL) < 0 ||
         sigaction(SIGPIPE, &sig_ign, NULL) < 0)
     {
@@ -641,14 +650,14 @@ main(int argc, char *const *argv)
             goto out;
     }
 
+    ret = EXIT_SUCCESS;
     while (likely(!aborted && (as_server || tll_length(wayl->terms) > 0))) {
-        if (unlikely(!fdm_poll(fdm)))
+        if (unlikely(!fdm_poll(fdm))) {
+            ret = foot_exit_failure;
             break;
+        }
     }
 
-    if (aborted || tll_length(wayl->terms) == 0)
-        ret = EXIT_SUCCESS;
-
 out:
     free(_cwd);
     server_destroy(server);
@@ -657,12 +666,13 @@ out:
     shm_fini();
     render_destroy(renderer);
     wayl_destroy(wayl);
+    key_binding_manager_destroy(key_binding_manager);
     reaper_destroy(reaper);
     fdm_signal_del(fdm, SIGTERM);
     fdm_signal_del(fdm, SIGINT);
     fdm_destroy(fdm);
 
-    config_free(conf);
+    config_free(&conf);
 
     if (unlink_pid_file)
         unlink(pid_file);
