@@ -252,7 +252,7 @@ osc_selection(struct terminal *term, char *string)
 
     LOG_DBG("clipboard: target = %s data = %s", string, p);
 
-    if (strlen(p) == 1 && p[0] == '?')
+    if (p[0] == '?' && p[1] == '\0')
         osc_from_clipboard(term, string);
     else
         osc_to_clipboard(term, string, p);
@@ -587,7 +587,7 @@ osc_dispatch(struct terminal *term)
             }
 
             /* Client queried for current value */
-            if (strlen(s_color) == 1 && s_color[0] == '?') {
+            if (s_color[0] == '?' && s_color[1] == '\0') {
                 uint32_t color = term->colors.table[idx];
                 uint8_t r = (color >> 16) & 0xff;
                 uint8_t g = (color >>  8) & 0xff;
@@ -596,8 +596,9 @@ osc_dispatch(struct terminal *term)
 
                 char reply[32];
                 size_t n = xsnprintf(
-                    reply, sizeof(reply), "\033]4;%u;rgb:%02x/%02x/%02x%s",
-                    idx, r, g, b, terminator);
+                    reply, sizeof(reply),
+                    "\033]4;%u;rgb:%02hhx%02hhx/%02hhx%02hhx/%02hhx%02hhx%s",
+                    idx, r, r, g, g, b, b, terminator);
                 term_to_slave(term, reply, n);
             }
 
@@ -614,7 +615,47 @@ osc_dispatch(struct terminal *term)
                         idx, term->colors.table[idx], color);
 
                 term->colors.table[idx] = color;
-                term_damage_view(term);
+
+                /* Dirty visible, affected cells */
+                for (int r = 0; r < term->rows; r++) {
+                    struct row *row = grid_row_in_view(term->grid, r);
+                    struct cell *cell = &row->cells[0];
+
+                    for (int c = 0; c < term->cols; c++, cell++) {
+                        bool dirty = false;
+
+                        switch (cell->attrs.fg_src) {
+                        case COLOR_BASE16:
+                        case COLOR_BASE256:
+                            if (cell->attrs.fg == idx)
+                                dirty = true;
+                            break;
+
+                        case COLOR_DEFAULT:
+                        case COLOR_RGB:
+                            /* Not affected */
+                            break;
+                        }
+
+                        switch (cell->attrs.bg_src) {
+                        case COLOR_BASE16:
+                        case COLOR_BASE256:
+                            if (cell->attrs.bg == idx)
+                                dirty = true;
+                            break;
+
+                        case COLOR_DEFAULT:
+                        case COLOR_RGB:
+                            /* Not affected */
+                            break;
+                        }
+
+                        if (dirty) {
+                            cell->attrs.clean = 0;
+                            row->dirty = true;
+                        }
+                    }
+                }
             }
         }
 
@@ -642,7 +683,7 @@ osc_dispatch(struct terminal *term)
         /* Set default foreground/background/highlight-bg/highlight-fg color */
 
         /* Client queried for current value */
-        if (strlen(string) == 1 && string[0] == '?') {
+        if (string[0] == '?' && string[1] == '\0') {
             uint32_t color = param == 10 ? term->colors.fg : term->colors.bg;
             uint8_t r = (color >> 16) & 0xff;
             uint8_t g = (color >>  8) & 0xff;
@@ -655,8 +696,9 @@ osc_dispatch(struct terminal *term)
              */
             char reply[32];
             size_t n = xsnprintf(
-                reply, sizeof(reply), "\033]%u;rgb:%02x/%02x/%02x%s",
-                param, r, g, b, terminator);
+                reply, sizeof(reply),
+                "\033]%u;rgb:%02hhx%02hhx/%02hhx%02hhx/%02hhx%02hhx%s",
+                param, r, r, g, g, b, b, terminator);
 
             term_to_slave(term, reply, n);
             break;
@@ -710,7 +752,7 @@ osc_dispatch(struct terminal *term)
     case 12: /* Set cursor color */
 
         /* Client queried for current value */
-        if (strlen(string) == 1 && string[0] == '?') {
+        if (string[0] == '?' && string[1] == '\0') {
             uint8_t r = (term->cursor_color.cursor >> 16) & 0xff;
             uint8_t g = (term->cursor_color.cursor >>  8) & 0xff;
             uint8_t b = (term->cursor_color.cursor >>  0) & 0xff;
@@ -744,6 +786,10 @@ osc_dispatch(struct terminal *term)
         term_damage_cursor(term);
         break;
 
+    case 22:  /* Set mouse cursor */
+        term_set_user_mouse_cursor(term, string);
+        break;
+
     case 30:  /* Set tab title */
         break;
 
@@ -754,7 +800,7 @@ osc_dispatch(struct terminal *term)
     case 104: {
         /* Reset Color Number 'c' (whole table if no parameter) */
 
-        if (strlen(string) == 0) {
+        if (string[0] == '\0') {
             LOG_DBG("resetting all colors");
             for (size_t i = 0; i < ALEN(term->colors.table); i++)
                 term->colors.table[i] = term->conf->colors.table[i];
@@ -823,6 +869,40 @@ osc_dispatch(struct terminal *term)
         term->colors.use_custom_selection = term->conf->colors.use_custom.selection;
         break;
 
+    case 133:
+        /*
+         * Shell integration; see
+         * https://iterm2.com/documentation-escape-codes.html (Shell
+         * Integration/FinalTerm)
+         *
+         * [PROMPT]prompt% [COMMAND_START] ls -l
+         * [COMMAND_EXECUTED]
+         * -rw-r--r-- 1 user group 127 May 1 2016 filename
+         * [COMMAND_FINISHED]
+         */
+        switch (string[0]) {
+        case 'A':
+            LOG_DBG("FTCS_PROMPT: %dx%d",
+                     term->grid->cursor.point.row,
+                    term->grid->cursor.point.col);
+
+            term->grid->cur_row->prompt_marker = true;
+            break;
+
+        case 'B':
+            LOG_DBG("FTCS_COMMAND_START");
+            break;
+
+        case 'C':
+            LOG_DBG("FTCS_COMMAND_EXECUTED");
+            break;
+
+        case 'D':
+            LOG_DBG("FTCS_COMMAND_FINISHED");
+            break;
+        }
+        break;
+
     case 555:
         osc_flash(term);
         break;
@@ -858,11 +938,20 @@ osc_dispatch(struct terminal *term)
 bool
 osc_ensure_size(struct terminal *term, size_t required_size)
 {
-    if (required_size <= term->vt.osc.size)
+    if (likely(required_size <= term->vt.osc.size))
         return true;
 
-    size_t new_size = (required_size + 127) / 128 * 128;
-    xassert(new_size > 0);
+    const size_t pow2_max = ~(SIZE_MAX >> 1);
+    if (unlikely(required_size > pow2_max)) {
+        LOG_ERR("required OSC buffer size (%zu) exceeds limit (%zu)",
+            required_size, pow2_max);
+        return false;
+    }
+
+    size_t new_size = max(term->vt.osc.size, 4096);
+    while (new_size < required_size) {
+        new_size <<= 1;
+    }
 
     uint8_t *new_data = realloc(term->vt.osc.data, new_size);
     if (new_data == NULL) {
@@ -870,6 +959,7 @@ osc_ensure_size(struct terminal *term, size_t required_size)
         return false;
     }
 
+    LOG_DBG("resized OSC buffer: %zu", new_size);
     term->vt.osc.data = new_data;
     term->vt.osc.size = new_size;
     return true;
