@@ -13,6 +13,7 @@
 
 #include "async.h"
 #include "config.h"
+#include "key-binding.h"
 #include "reaper.h"
 #include "sixel.h"
 #include "user-notification.h"
@@ -68,6 +69,12 @@ void render_refresh(struct terminal *term) {}
 void render_refresh_csd(struct terminal *term) {}
 void render_refresh_title(struct terminal *term) {}
 
+bool
+render_xcursor_is_valid(const struct seat *seat, const char *cursor)
+{
+    return true;
+}
+
 bool
 render_xcursor_set(struct seat *seat, struct terminal *term, const char *xcursor)
 {
@@ -91,14 +98,16 @@ bool wayl_win_set_urgent(struct wl_window *win) { return true; }
 
 bool
 spawn(struct reaper *reaper, const char *cwd, char *const argv[],
-      int stdin_fd, int stdout_fd, int stderr_fd)
+      int stdin_fd, int stdout_fd, int stderr_fd,
+      const char *xdg_activation_token)
 {
     return true;
 }
 
 pid_t
 slave_spawn(
-    int ptmx, int argc, const char *cwd, char *const *argv, const char *term_env,
+    int ptmx, int argc, const char *cwd, char *const *argv, char *const *envp,
+    const env_var_list_t *extra_env_vars, const char *term_env,
     const char *conf_shell, bool login_shell,
     const user_notifications_t *notifications)
 {
@@ -164,6 +173,37 @@ void get_current_modifiers(const struct seat *seat,
                            xkb_mod_mask_t *effective,
                            xkb_mod_mask_t *consumed, uint32_t key) {}
 
+static struct key_binding_set kbd;
+static bool kbd_initialized = false;
+
+struct key_binding_set *
+key_binding_for(
+    struct key_binding_manager *mgr, const struct terminal *term,
+    const struct seat *seat)
+{
+    return &kbd;
+}
+
+void
+key_binding_new_for_term(
+    struct key_binding_manager *mgr, const struct terminal *term)
+{
+    if (!kbd_initialized) {
+        kbd_initialized = true;
+        kbd = (struct key_binding_set){
+            .key = tll_init(),
+            .search = tll_init(),
+            .url = tll_init(),
+            .mouse = tll_init(),
+            .selection_overrides = 0,
+        };
+    }
+}
+
+void
+key_binding_unref_term(struct key_binding_manager *mgr, const struct terminal *term)
+{
+}
 
 int
 main(int argc, const char *const *argv)
@@ -234,8 +274,10 @@ main(int argc, const char *const *argv)
             .end = row_count,
         },
         .selection = {
-            .start = {-1, -1},
-            .end = {-1, -1},
+            .coords = {
+                .start = {-1, -1},
+                .end = {-1, -1},
+            },
         },
         .delayed_render_timer = {
             .lower_fd = lower_fd,
