@@ -13,10 +13,12 @@
 #include "extract.h"
 #include "grid.h"
 #include "input.h"
+#include "key-binding.h"
 #include "misc.h"
 #include "render.h"
 #include "selection.h"
 #include "shm.h"
+#include "unicode-mode.h"
 #include "util.h"
 #include "xmalloc.h"
 
@@ -34,25 +36,26 @@
 static int
 ensure_view_is_allocated(struct terminal *term, int new_view)
 {
-    int view_end = (new_view + term->rows - 1) & (term->grid->num_rows - 1);
+    struct grid *grid = term->grid;
+    int view_end = (new_view + term->rows - 1) & (grid->num_rows - 1);
 
-    if (term->grid->rows[new_view] == NULL) {
-        while (term->grid->rows[new_view] == NULL)
-            new_view = (new_view + 1) & (term->grid->num_rows - 1);
+    if (grid->rows[new_view] == NULL) {
+        while (grid->rows[new_view] == NULL)
+            new_view = (new_view + 1) & (grid->num_rows - 1);
     }
 
-    else if (term->grid->rows[view_end] == NULL) {
-        while (term->grid->rows[view_end] == NULL) {
+    else if (grid->rows[view_end] == NULL) {
+        while (grid->rows[view_end] == NULL) {
             new_view--;
             if (new_view < 0)
-                new_view += term->grid->num_rows;
-            view_end = (new_view + term->rows - 1) & (term->grid->num_rows - 1);
+                new_view += grid->num_rows;
+            view_end = (new_view + term->rows - 1) & (grid->num_rows - 1);
         }
     }
 
 #if defined(_DEBUG)
     for (size_t r = 0; r < term->rows; r++)
-        xassert(term->grid->rows[(new_view + r) & (term->grid->num_rows - 1)] != NULL);
+        xassert(grid->rows[(new_view + r) & (grid->num_rows - 1)] != NULL);
 #endif
 
     return new_view;
@@ -80,10 +83,7 @@ search_ensure_size(struct terminal *term, size_t wanted_size)
 static bool
 has_wrapped_around(const struct terminal *term, int abs_row_no)
 {
-    int scrollback_start = term->grid->offset + term->rows;
-    int rebased_row = abs_row_no - scrollback_start + term->grid->num_rows;
-    rebased_row &= term->grid->num_rows - 1;
-
+    int rebased_row = grid_row_abs_to_sb(term->grid, term->rows, abs_row_no);
     return rebased_row == 0;
 }
 
@@ -117,6 +117,11 @@ search_cancel_keep_selection(struct terminal *term)
 
     term_xcursor_update(term);
     render_refresh(term);
+
+    /* Work around Sway bug - unmapping a sub-surface does not damage
+     * the underlying surface */
+    term_damage_margins(term);
+    term_damage_view(term);
 }
 
 void
@@ -134,11 +139,13 @@ search_begin(struct terminal *term)
     }
 
     /* On-demand instantiate wayland surface */
-    bool ret = wayl_win_subsurface_new(term->window, &term->window->search);
+    bool ret = wayl_win_subsurface_new(
+        term->window, &term->window->search, false);
     xassert(ret);
 
-    term->search.original_view = term->grid->view;
-    term->search.view_followed_offset = term->grid->view == term->grid->offset;
+    const struct grid *grid = term->grid;
+    term->search.original_view = grid->view;
+    term->search.view_followed_offset = grid->view == grid->offset;
     term->is_searching = true;
 
     term->search.len = 0;
@@ -169,83 +176,82 @@ search_selection_cancelled(struct terminal *term)
 }
 
 static void
-search_update_selection(struct terminal *term,
-                        int start_row, int start_col,
-                        int end_row, int end_col)
+search_update_selection(struct terminal *term, const struct range *match)
 {
+    struct grid *grid = term->grid;
+    int start_row = match->start.row;
+    int start_col = match->start.col;
+    int end_row = match->end.row;
+    int end_col = match->end.col;
+
+    xassert(start_row >= 0);
+    xassert(start_row < grid->num_rows);
+
     bool move_viewport = true;
 
-    int view_end = (term->grid->view + term->rows - 1) & (term->grid->num_rows - 1);
-    if (view_end >= term->grid->view) {
+    int view_end = (grid->view + term->rows - 1) & (grid->num_rows - 1);
+    if (view_end >= grid->view) {
         /* Viewport does *not* wrap around */
-        if (start_row >= term->grid->view && end_row <= view_end)
+        if (start_row >= grid->view && end_row <= view_end)
             move_viewport = false;
     } else {
         /* Viewport wraps */
-        if (start_row >= term->grid->view || end_row <= view_end)
+        if (start_row >= grid->view || end_row <= view_end)
             move_viewport = false;
     }
 
     if (move_viewport) {
-        int old_view = term->grid->view;
-        int new_view = start_row - term->rows / 2;
+        int rebased_new_view = grid_row_abs_to_sb(grid, term->rows, start_row);
 
-        while (new_view < 0)
-            new_view += term->grid->num_rows;
+        rebased_new_view -= term->rows / 2;
+        rebased_new_view =
+            min(max(rebased_new_view, 0), grid->num_rows - term->rows);
 
-        new_view = ensure_view_is_allocated(term, new_view);
+        const int old_view = grid->view;
+        int new_view = grid_row_sb_to_abs(grid, term->rows, rebased_new_view);
 
-        /* Don't scroll past scrollback history */
-        int end = (term->grid->offset + term->rows - 1) & (term->grid->num_rows - 1);
-        if (end >= term->grid->offset) {
-            /* Not wrapped */
-            if (new_view >= term->grid->offset && new_view <= end)
-                new_view = term->grid->offset;
-        } else {
-            if (new_view >= term->grid->offset || new_view <= end)
-                new_view = term->grid->offset;
+        /* Scrollback may not be completely filled yet */
+        {
+            const int mask = grid->num_rows - 1;
+            while (grid->rows[new_view] == NULL)
+                new_view = (new_view + 1) & mask;
         }
 
 #if defined(_DEBUG)
         /* Verify all to-be-visible rows have been allocated */
         for (int r = 0; r < term->rows; r++)
-            xassert(term->grid->rows[(new_view + r) & (term->grid->num_rows - 1)] != NULL);
+            xassert(grid->rows[(new_view + r) & (grid->num_rows - 1)] != NULL);
+#endif
+
+#if defined(_DEBUG)
+        {
+            int rel_start_row = grid_row_abs_to_sb(grid, term->rows, start_row);
+            int rel_view = grid_row_abs_to_sb(grid, term->rows, new_view);
+            xassert(rel_view <= rel_start_row);
+            xassert(rel_start_row < rel_view + term->rows);
+        }
 #endif
 
         /* Update view */
-        term->grid->view = new_view;
+        grid->view = new_view;
         if (new_view != old_view)
             term_damage_view(term);
     }
 
-    /* Selection endpoint is inclusive */
-    if (--end_col < 0) {
-        end_col = term->cols - 1;
-        end_row--;
-    }
-
-    /* Begin a new selection if the start coords changed */
     if (start_row != term->search.match.row ||
         start_col != term->search.match.col)
     {
-        int selection_row = start_row - term->grid->view;
-        while (selection_row < 0)
-            selection_row += term->grid->num_rows;
+        int selection_row = start_row - grid->view + grid->num_rows;
+        selection_row &= grid->num_rows - 1;
 
-        xassert(selection_row >= 0 &&
-               selection_row < term->grid->num_rows);
         selection_start(
             term, start_col, selection_row, SELECTION_CHAR_WISE, false);
     }
 
     /* Update selection endpoint */
     {
-        int selection_row = end_row - term->grid->view;
-        while (selection_row < 0)
-            selection_row += term->grid->num_rows;
-
-        xassert(selection_row >= 0 &&
-               selection_row < term->grid->num_rows);
+        int selection_row = end_row - grid->view + grid->num_rows;
+        selection_row &= grid->num_rows - 1;
         selection_update(term, end_col, selection_row);
     }
 }
@@ -283,11 +289,127 @@ matches_cell(const struct terminal *term, const struct cell *cell, size_t search
     return composed != NULL ? 1 + composed->count : 1;
 }
 
-static void
-search_find_next(struct terminal *term)
+static bool
+find_next(struct terminal *term, enum search_direction direction,
+          struct coord abs_start, struct coord abs_end, struct range *match)
 {
-    bool backward = term->search.direction == SEARCH_BACKWARD;
-    term->search.direction = SEARCH_BACKWARD;
+#define ROW_DEC(_r) ((_r) = ((_r) - 1 + grid->num_rows) & (grid->num_rows - 1))
+#define ROW_INC(_r) ((_r) = ((_r) + 1) & (grid->num_rows - 1))
+
+    struct grid *grid = term->grid;
+    const bool backward = direction != SEARCH_FORWARD;
+
+    LOG_DBG("%s: start: %dx%d, end: %dx%d", backward ? "backward" : "forward",
+            abs_start.row, abs_start.col, abs_end.row, abs_end.col);
+
+    xassert(abs_start.row >= 0);
+    xassert(abs_start.row < grid->num_rows);
+    xassert(abs_start.col >= 0);
+    xassert(abs_start.col < term->cols);
+
+    xassert(abs_end.row >= 0);
+    xassert(abs_end.row < grid->num_rows);
+    xassert(abs_end.col >= 0);
+    xassert(abs_end.col < term->cols);
+
+    for (int match_start_row = abs_start.row, match_start_col = abs_start.col;
+         ;
+         backward ? ROW_DEC(match_start_row) : ROW_INC(match_start_row)) {
+
+        const struct row *row = grid->rows[match_start_row];
+        if (row == NULL) {
+            if (match_start_row == abs_end.row)
+                break;
+            continue;
+        }
+
+        for (;
+             backward ? match_start_col >= 0 : match_start_col < term->cols;
+             backward ? match_start_col-- : match_start_col++)
+        {
+            if (matches_cell(term, &row->cells[match_start_col], 0) < 0) {
+                if (match_start_row == abs_end.row &&
+                    match_start_col == abs_end.col)
+                {
+                    break;
+                }
+                continue;
+            }
+
+            /*
+             * Got a match on the first letter. Now we'll see if the
+             * rest of the search buffer matches.
+             */
+
+            LOG_DBG("search: initial match at row=%d, col=%d",
+                    match_start_row, match_start_col);
+
+            int match_end_row = match_start_row;
+            int match_end_col = match_start_col;
+            const struct row *match_row = row;
+            size_t match_len = 0;
+
+            for (size_t i = 0; i < term->search.len;) {
+                if (match_end_col >= term->cols) {
+                    ROW_INC(match_end_row);
+                    match_end_col = 0;
+
+                    match_row = grid->rows[match_end_row];
+                    if (match_row == NULL)
+                        break;
+                }
+
+                if (match_row->cells[match_end_col].wc >= CELL_SPACER) {
+                    match_end_col++;
+                    continue;
+                }
+
+                ssize_t additional_chars = matches_cell(
+                    term, &match_row->cells[match_end_col], i);
+                if (additional_chars < 0)
+                    break;
+
+                i += additional_chars;
+                match_len += additional_chars;
+                match_end_col++;
+
+                while (match_row->cells[match_end_col].wc > CELL_SPACER)
+                    match_end_col++;
+            }
+
+            if (match_len != term->search.len) {
+                /* Didn't match (completely) */
+
+                if (match_start_row == abs_end.row &&
+                    match_start_col == abs_end.col)
+                {
+                    break;
+                }
+
+                continue;
+            }
+
+            *match = (struct range){
+                .start = {match_start_col, match_start_row},
+                .end = {match_end_col - 1, match_end_row},
+            };
+
+            return true;
+        }
+
+        if (match_start_row == abs_end.row && match_start_col == abs_end.col)
+            break;
+
+        match_start_col = backward ? term->cols - 1 : 0;
+    }
+
+    return false;
+}
+
+static void
+search_find_next(struct terminal *term, enum search_direction direction)
+{
+    struct grid *grid = term->grid;
 
     if (term->search.len == 0) {
         term->search.match = (struct coord){-1, -1};
@@ -296,113 +418,193 @@ search_find_next(struct terminal *term)
         return;
     }
 
-    int start_row = term->search.match.row;
-    int start_col = term->search.match.col;
+    struct coord start = term->search.match;
     size_t len = term->search.match_len;
 
-    xassert((len == 0 && start_row == -1 && start_col == -1) ||
-           (len > 0 && start_row >= 0 && start_col >= 0));
+    xassert((len == 0 && start.row == -1 && start.col == -1) ||
+           (len > 0 && start.row >= 0 && start.col >= 0));
 
     if (len == 0) {
-        if (backward) {
-            start_row = grid_row_absolute_in_view(term->grid, term->rows - 1);
-            start_col = term->cols - 1;
-        } else {
-            start_row = grid_row_absolute_in_view(term->grid, 0);
-            start_col = 0;
+        /* No previous match, start from the top, or bottom, of the scrollback */
+        switch (direction) {
+        case SEARCH_FORWARD:
+            start.row = grid_row_absolute_in_view(grid, 0);
+            start.col = 0;
+            break;
+
+        case SEARCH_BACKWARD:
+        case SEARCH_BACKWARD_SAME_POSITION:
+            start.row = grid_row_absolute_in_view(grid, term->rows - 1);
+            start.col = term->cols - 1;
+            break;
         }
-    }
+    } else {
+        /* Continue from last match */
+        xassert(start.row >= 0);
+        xassert(start.col >= 0);
 
-    LOG_DBG("search: update: %s: starting at row=%d col=%d (offset = %d, view = %d)",
-            backward ? "backward" : "forward", start_row, start_col,
-            term->grid->offset, term->grid->view);
+        switch (direction) {
+        case SEARCH_BACKWARD_SAME_POSITION:
+            break;
 
-#define ROW_DEC(_r) ((_r) = ((_r) - 1 + term->grid->num_rows) & (term->grid->num_rows - 1))
-#define ROW_INC(_r) ((_r) = ((_r) + 1) & (term->grid->num_rows - 1))
-
-    /* Scan backward from current end-of-output */
-    /* TODO: don't search "scrollback" in alt screen? */
-    for (size_t r = 0;
-         r < term->grid->num_rows;
-         backward ? ROW_DEC(start_row) : ROW_INC(start_row), r++)
-    {
-        for (;
-             backward ? start_col >= 0 : start_col < term->cols;
-             backward ? start_col-- : start_col++)
-        {
-            const struct row *row = term->grid->rows[start_row];
-            if (row == NULL)
-                continue;
-
-            if (matches_cell(term, &row->cells[start_col], 0) < 0)
-                continue;
-
-            /*
-             * Got a match on the first letter. Now we'll see if the
-             * rest of the search buffer matches.
-             */
-
-            LOG_DBG("search: initial match at row=%d, col=%d", start_row, start_col);
-
-            int end_row = start_row;
-            int end_col = start_col;
-            size_t match_len = 0;
-
-            for (size_t i = 0; i < term->search.len;) {
-                if (end_col >= term->cols) {
-                    end_row = (end_row + 1) & (term->grid->num_rows - 1);
-                    end_col = 0;
-
-                    if (has_wrapped_around(term, end_row))
-                        break;
-
-                    row = term->grid->rows[end_row];
-                }
-
-                if (row->cells[end_col].wc >= CELL_SPACER) {
-                    end_col++;
-                    continue;
-                }
-
-                ssize_t additional_chars = matches_cell(term, &row->cells[end_col], i);
-                if (additional_chars < 0)
-                    break;
-
-                i += additional_chars;
-                match_len += additional_chars;
-                end_col++;
+        case SEARCH_BACKWARD:
+            if (--start.col < 0) {
+                start.col = term->cols - 1;
+                start.row += grid->num_rows - 1;
+                start.row &= grid->num_rows - 1;
             }
+            break;
 
-            if (match_len != term->search.len) {
-                /* Didn't match (completely) */
-                continue;
+        case SEARCH_FORWARD:
+            if (++start.col >= term->cols) {
+                start.col = 0;
+                start.row++;
+                start.row &= grid->num_rows - 1;
             }
-
-            /*
-             * We matched the entire buffer. Move view to ensure the
-             * match is visible, create a selection and return.
-             */
-            search_update_selection(term, start_row, start_col, end_row, end_col);
-
-            /* Update match state */
-            term->search.match.row = start_row;
-            term->search.match.col = start_col;
-            term->search.match_len = match_len;
-
-            return;
+            break;
         }
 
-        start_col = backward ? term->cols - 1 : 0;
+        xassert(start.row >= 0);
+        xassert(start.row < grid->num_rows);
+        xassert(start.col >= 0);
+        xassert(start.col < term->cols);
     }
 
-    /* No match */
-    LOG_DBG("no match");
-    term->search.match = (struct coord){-1, -1};
-    term->search.match_len = 0;
-    selection_cancel(term);
+    LOG_DBG(
+        "update: %s: starting at row=%d col=%d "
+        "(offset = %d, view = %d)",
+        direction != SEARCH_FORWARD ? "backward" : "forward",
+        start.row, start.col,
+        grid->offset, grid->view);
+
+    struct coord end = start;
+    switch (direction) {
+    case SEARCH_FORWARD:
+        /* Search forward, until we reach the cell *before* current start */
+        if (--end.col < 0) {
+            end.col = term->cols - 1;
+            end.row += grid->num_rows - 1;
+            end.row &= grid->num_rows - 1;
+        }
+        break;
+
+    case SEARCH_BACKWARD:
+    case SEARCH_BACKWARD_SAME_POSITION:
+        /* Search backwards, until we reach the cell *after* current start */
+        if (++end.col >= term->cols) {
+            end.col = 0;
+            end.row++;
+            end.row &= grid->num_rows - 1;
+        }
+        break;
+    }
+
+    struct range match;
+    bool found = find_next(term, direction, start, end, &match);
+
+    if (found) {
+        LOG_DBG("primary match found at %dx%d",
+                match.start.row, match.start.col);
+        search_update_selection(term, &match);
+        term->search.match = match.start;
+        term->search.match_len = term->search.len;
+    } else {
+        LOG_DBG("no match");
+        term->search.match = (struct coord){-1, -1};
+        term->search.match_len = 0;
+        selection_cancel(term);
+    }
 #undef ROW_DEC
 }
 
+struct search_match_iterator
+search_matches_new_iter(struct terminal *term)
+{
+    return (struct search_match_iterator){
+        .term = term,
+        .start = {0, 0},
+    };
+}
+
+struct range
+search_matches_next(struct search_match_iterator *iter)
+{
+    struct terminal *term = iter->term;
+    struct grid *grid = term->grid;
+
+    if (term->search.match_len == 0)
+        goto no_match;
+
+    if (iter->start.row >= term->rows)
+        goto no_match;
+
+    xassert(iter->start.row >= 0);
+    xassert(iter->start.row < term->rows);
+    xassert(iter->start.col >= 0);
+    xassert(iter->start.col < term->cols);
+
+    struct coord abs_start = iter->start;
+    abs_start.row = grid_row_absolute_in_view(grid, abs_start.row);
+
+    struct coord abs_end = {
+        term->cols - 1,
+        grid_row_absolute_in_view(grid, term->rows - 1)};
+
+    struct range match;
+    bool found = find_next(term, SEARCH_FORWARD, abs_start, abs_end, &match);
+    if (!found)
+        goto no_match;
+
+    LOG_DBG("match at (absolute coordinates) %dx%d-%dx%d",
+            match.start.row, match.start.col,
+            match.end.row, match.end.col);
+
+    /* Convert absolute row numbers back to view relative */
+    match.start.row = match.start.row - grid->view + grid->num_rows;
+    match.start.row &= grid->num_rows - 1;
+    match.end.row = match.end.row - grid->view + grid->num_rows;
+    match.end.row &= grid->num_rows - 1;
+
+    LOG_DBG("match at (view-local coordinates) %dx%d-%dx%d, view=%d",
+            match.start.row, match.start.col,
+            match.end.row, match.end.col, grid->view);
+
+    xassert(match.start.row >= 0);
+    xassert(match.start.row < term->rows);
+    xassert(match.end.row >= 0);
+    xassert(match.end.row < term->rows);
+
+    /* Assert match end comes *after* the match start */
+    xassert(match.end.row > match.start.row ||
+            (match.end.row == match.start.row &&
+             match.end.col >= match.start.col));
+
+    /* Assert the match starts at, or after, the iterator position */
+    xassert(match.start.row > iter->start.row ||
+            (match.start.row == iter->start.row &&
+             match.start.col >= iter->start.col));
+
+    /* Continue at next column, next time */
+    iter->start.row = match.start.row;
+    iter->start.col = match.start.col + 1;
+
+    if (iter->start.col >= term->cols) {
+        iter->start.col = 0;
+        iter->start.row++;  /* Overflow is caught in next iteration */
+    }
+
+    xassert(iter->start.row >= 0);
+    xassert(iter->start.row <= term->rows);
+    xassert(iter->start.col >= 0);
+    xassert(iter->start.col < term->cols);
+    return match;
+
+no_match:
+    iter->start.row = -1;
+    iter->start.col = -1;
+    return (struct range){{-1, -1}, {-1,  -1}};
+}
+
 static void
 add_wchars(struct terminal *term, char32_t *src, size_t count)
 {
@@ -450,47 +652,57 @@ search_match_to_end_of_word(struct terminal *term, bool spaces_only)
     if (term->search.match_len == 0)
         return;
 
-    xassert(term->selection.end.row != -1);
+    xassert(term->selection.coords.end.row >= 0);
 
+    struct grid *grid = term->grid;
     const bool move_cursor = term->search.cursor == term->search.len;
 
-    const struct coord old_end = term->selection.end;
+    struct coord old_end = selection_get_end(term);
     struct coord new_end = old_end;
     struct row *row = NULL;
 
-#define newline(coord) __extension__                                    \
+    xassert(new_end.row >= 0);
+    xassert(new_end.row < grid->num_rows);
+
+    /* Advances a coordinate by one column, to the right. Returns
+     * false if we’ve reached the scrollback wrap-around */
+#define advance_pos(coord) __extension__                                \
         ({                                                              \
             bool wrapped_around = false;                                \
-            if (++(coord).col >= term->cols) {                           \
-                (coord).row = ((coord).row + 1) & (term->grid->num_rows - 1); \
+            if (++(coord).col >= term->cols) {                          \
+                (coord).row = ((coord).row + 1) & (grid->num_rows - 1); \
                 (coord).col = 0;                                        \
-                row = term->grid->rows[(coord).row];                    \
+                row = grid->rows[(coord).row];                          \
                 if (has_wrapped_around(term, (coord.row)))              \
                     wrapped_around = true;                              \
             }                                                           \
-            wrapped_around;                                             \
+            !wrapped_around;                                             \
         })
 
     /* First character to consider is the *next* character */
-    if (newline(new_end))
+    if (!advance_pos(new_end))
         return;
 
-    xassert(term->grid->rows[new_end.row] != NULL);
+    xassert(new_end.row >= 0);
+    xassert(new_end.row < grid->num_rows);
+    xassert(grid->rows[new_end.row] != NULL);
 
     /* Find next word boundary */
-    new_end.row -= term->grid->view;
-    selection_find_word_boundary_right(term, &new_end, spaces_only);
-    new_end.row += term->grid->view;
+    new_end.row -= grid->view + grid->num_rows;
+    new_end.row &= grid->num_rows - 1;
+    selection_find_word_boundary_right(term, &new_end, spaces_only, false);
+    new_end.row += grid->view;
+    new_end.row &= grid->num_rows - 1;
 
     struct coord pos = old_end;
-    row = term->grid->rows[pos.row];
+    row = grid->rows[pos.row];
 
     struct extraction_context *ctx = extract_begin(SELECTION_NONE, false);
     if (ctx == NULL)
         return;
 
     do {
-        if (newline(pos))
+        if (!advance_pos(pos))
             break;
         if (!extract_one(term, row, &row->cells[pos.col], pos.col, ctx))
             break;
@@ -520,15 +732,12 @@ search_match_to_end_of_word(struct terminal *term, bool spaces_only)
     if (move_cursor)
         term->search.cursor = term->search.len;
 
-    /* search_update_selection() expected end coordinate to be *exclusive* */
-    newline(new_end);
-    search_update_selection(
-        term, term->search.match.row, term->search.match.col,
-        new_end.row, new_end.col);
+    struct range match = {.start = term->search.match, .end = new_end};
+    search_update_selection(term, &match);
 
     term->search.match_len = term->search.len;
 
-#undef newline
+#undef advance_pos
 }
 
 static size_t
@@ -598,16 +807,20 @@ from_clipboard_done(void *user)
     struct terminal *term = user;
 
     LOG_DBG("search: buffer: %ls", (const wchar_t *)term->search.buf);
-    search_find_next(term);
+    search_find_next(term, SEARCH_BACKWARD_SAME_POSITION);
     render_refresh_search(term);
 }
 
 static bool
 execute_binding(struct seat *seat, struct terminal *term,
-                enum bind_action_search action, uint32_t serial,
-                bool *update_search_result, bool *redraw)
+                const struct key_binding *binding, uint32_t serial,
+                bool *update_search_result, enum search_direction *direction,
+                bool *redraw)
 {
     *update_search_result = *redraw = false;
+    const enum bind_action_search action = binding->action;
+
+    struct grid *grid = term->grid;
 
     switch (action) {
     case BIND_ACTION_SEARCH_NONE:
@@ -615,12 +828,11 @@ execute_binding(struct seat *seat, struct terminal *term,
 
     case BIND_ACTION_SEARCH_CANCEL:
         if (term->search.view_followed_offset)
-            term->grid->view = term->grid->offset;
+            grid->view = grid->offset;
         else {
-            term->grid->view = ensure_view_is_allocated(
+            grid->view = ensure_view_is_allocated(
                 term, term->search.original_view);
         }
-        term_damage_view(term);
         search_cancel(term);
         return true;
 
@@ -638,20 +850,7 @@ execute_binding(struct seat *seat, struct terminal *term,
             term->search.last.len = 0;
         }
 
-        else if (term->search.match_len > 0) {
-            int new_col = term->search.match.col - 1;
-            int new_row = term->search.match.row;
-
-            if (new_col < 0) {
-                new_col = term->cols - 1;
-                new_row--;
-            }
-
-            if (new_row >= 0) {
-                term->search.match.col = new_col;
-                term->search.match.row = new_row;
-            }
-        }
+        *direction = SEARCH_BACKWARD;
         *update_search_result = *redraw = true;
         return true;
 
@@ -664,21 +863,7 @@ execute_binding(struct seat *seat, struct terminal *term,
             term->search.last.len = 0;
         }
 
-        else if (term->search.match_len > 0) {
-            int new_col = term->search.match.col + 1;
-            int new_row = term->search.match.row;
-
-            if (new_col >= term->cols) {
-                new_col = 0;
-                new_row++;
-            }
-
-            if (new_row < term->grid->num_rows) {
-                term->search.match.col = new_col;
-                term->search.match.row = new_row;
-                term->search.direction = SEARCH_FORWARD;
-             }
-        }
+        *direction = SEARCH_FORWARD;
         *update_search_result = *redraw = true;
         return true;
 
@@ -809,6 +994,10 @@ execute_binding(struct seat *seat, struct terminal *term,
         *update_search_result = *redraw = true;
         return true;
 
+    case BIND_ACTION_SEARCH_UNICODE_INPUT:
+        unicode_mode_activate(seat);
+        return true;
+
     case BIND_ACTION_SEARCH_COUNT:
         BUG("Invalid action type");
         return true;
@@ -819,44 +1008,54 @@ execute_binding(struct seat *seat, struct terminal *term,
 }
 
 void
-search_input(struct seat *seat, struct terminal *term, uint32_t key,
+search_input(struct seat *seat, struct terminal *term,
+             const struct key_binding_set *bindings, uint32_t key,
              xkb_keysym_t sym, xkb_mod_mask_t mods, xkb_mod_mask_t consumed,
+             xkb_mod_mask_t locked,
              const xkb_keysym_t *raw_syms, size_t raw_count,
              uint32_t serial)
 {
-    LOG_DBG("search: input: sym=%d/0x%x, mods=0x%08x", sym, sym, mods);
+    LOG_DBG("search: input: sym=%d/0x%x, mods=0x%08x, consumed=0x%08x",
+            sym, sym, mods, consumed);
 
+    const xkb_mod_mask_t bind_mods =
+        mods & seat->kbd.bind_significant & ~locked;
+    const xkb_mod_mask_t bind_consumed =
+        consumed & seat->kbd.bind_significant & ~locked;
     enum xkb_compose_status compose_status = seat->kbd.xkb_compose_state != NULL
       ? xkb_compose_state_get_status(seat->kbd.xkb_compose_state)
       : XKB_COMPOSE_NOTHING;
 
+    enum search_direction search_direction = SEARCH_BACKWARD_SAME_POSITION;
     bool update_search_result = false;
     bool redraw = false;
 
     /* Key bindings */
-    tll_foreach(seat->kbd.bindings.search, it) {
+    tll_foreach(bindings->search, it) {
         const struct key_binding *bind = &it->item;
 
         /* Match translated symbol */
-        if (bind->sym == sym &&
-            bind->mods == (mods & ~consumed)) {
+        if (bind->k.sym == sym &&
+            bind->mods == (bind_mods & ~bind_consumed)) {
 
-            if (execute_binding(seat, term, bind->action, serial,
-                                &update_search_result, &redraw))
+            if (execute_binding(seat, term, bind, serial,
+                                &update_search_result, &search_direction,
+                                &redraw))
             {
                 goto update_search;
             }
             return;
         }
 
-        if (bind->mods != mods)
+        if (bind->mods != bind_mods || bind_mods != (mods & ~locked))
             continue;
 
         /* Match untranslated symbols */
         for (size_t i = 0; i < raw_count; i++) {
-            if (bind->sym == raw_syms[i]) {
-                if (execute_binding(seat, term, bind->action, serial,
-                                    &update_search_result, &redraw))
+            if (bind->k.sym == raw_syms[i]) {
+                if (execute_binding(seat, term, bind, serial,
+                                    &update_search_result, &search_direction,
+                                    &redraw))
                 {
                     goto update_search;
                 }
@@ -865,10 +1064,11 @@ search_input(struct seat *seat, struct terminal *term, uint32_t key,
         }
 
         /* Match raw key code */
-        tll_foreach(bind->key_codes, code) {
+        tll_foreach(bind->k.key_codes, code) {
             if (code->item == key) {
-                if (execute_binding(seat, term, bind->action, serial,
-                                    &update_search_result, &redraw))
+                if (execute_binding(seat, term, bind, serial,
+                                    &update_search_result, &search_direction,
+                                    &redraw))
                 {
                     goto update_search;
                 }
@@ -892,6 +1092,7 @@ search_input(struct seat *seat, struct terminal *term, uint32_t key,
     }
 
     update_search_result = redraw = count > 0;
+    search_direction = SEARCH_BACKWARD_SAME_POSITION;
 
     if (count == 0)
         return;
@@ -901,7 +1102,7 @@ search_input(struct seat *seat, struct terminal *term, uint32_t key,
 update_search:
     LOG_DBG("search: buffer: %ls", (const wchar_t *)term->search.buf);
     if (update_search_result)
-        search_find_next(term);
+        search_find_next(term, search_direction);
     if (redraw)
         render_refresh_search(term);
 }
