@@ -14,10 +14,11 @@
 #include <tllist.h>
 #include <fcft/fcft.h>
 
-//#include "config.h"
 #include "composed.h"
+#include "config.h"
 #include "debug.h"
 #include "fdm.h"
+#include "key-binding.h"
 #include "macros.h"
 #include "reaper.h"
 #include "shm.h"
@@ -79,6 +80,11 @@ struct coord {
     int row;
 };
 
+struct range {
+    struct coord start;
+    struct coord end;
+};
+
 struct cursor {
     struct coord point;
     bool lcf;
@@ -110,9 +116,13 @@ struct row_data {
 
 struct row {
     struct cell *cells;
+    struct row_data *extra;
+
     bool dirty;
     bool linebreak;
-    struct row_data *extra;
+
+    /* Shell integration */
+    bool prompt_marker;
 };
 
 struct sixel {
@@ -248,8 +258,6 @@ enum mouse_reporting {
     MOUSE_SGR_PIXELS,    /* ?1016h */
 };
 
-enum cursor_style { CURSOR_BLOCK, CURSOR_UNDERLINE, CURSOR_BEAM };
-
 enum selection_kind {
     SELECTION_NONE,
     SELECTION_CHAR_WISE,
@@ -259,6 +267,7 @@ enum selection_kind {
 };
 enum selection_direction {SELECTION_UNDIR, SELECTION_LEFT, SELECTION_RIGHT};
 enum selection_scroll_direction {SELECTION_SCROLL_NOT, SELECTION_SCROLL_UP, SELECTION_SCROLL_DOWN};
+enum search_direction { SEARCH_BACKWARD_SAME_POSITION, SEARCH_BACKWARD, SEARCH_FORWARD };
 
 struct ptmx_buffer {
     void *data;
@@ -269,10 +278,6 @@ struct ptmx_buffer {
 enum term_surface {
     TERM_SURF_NONE,
     TERM_SURF_GRID,
-    TERM_SURF_SEARCH,
-    TERM_SURF_SCROLLBACK_INDICATOR,
-    TERM_SURF_RENDER_TIMER,
-    TERM_SURF_JUMP_LABEL,
     TERM_SURF_TITLE,
     TERM_SURF_BORDER_LEFT,
     TERM_SURF_BORDER_RIGHT,
@@ -283,15 +288,21 @@ enum term_surface {
     TERM_SURF_BUTTON_CLOSE,
 };
 
+enum overlay_style {
+    OVERLAY_NONE,
+    OVERLAY_SEARCH,
+    OVERLAY_FLASH,
+    OVERLAY_UNICODE_MODE,
+};
+
 typedef tll(struct ptmx_buffer) ptmx_buffer_list_t;
 
-enum url_action { URL_ACTION_COPY, URL_ACTION_LAUNCH };
+enum url_action { URL_ACTION_COPY, URL_ACTION_LAUNCH, URL_ACTION_PERSISTENT };
 struct url {
     uint64_t id;
     char *url;
     char32_t *key;
-    struct coord start;
-    struct coord end;
+    struct range range;
     enum url_action action;
     bool url_mode_dont_change_url_attr; /* Entering/exiting URL mode doesn’t touch the cells’ attr.url */
     bool osc8;
@@ -299,12 +310,6 @@ struct url {
 };
 typedef tll(struct url) url_list_t;
 
-/* If px != 0 then px is valid, otherwise pt is valid */
-struct pt_or_px {
-    int16_t px;
-    float pt;
-};
-
 struct terminal {
     struct fdm *fdm;
     struct reaper *reaper;
@@ -335,13 +340,13 @@ struct terminal {
     bool bracketed_paste;
     bool focus_events;
     bool alt_scrolling;
-    bool modify_escape_key;
     bool modify_other_keys_2;  /* True when modifyOtherKeys=2 (i.e. “CSI >4;2m”) */
     enum cursor_origin origin;
     enum cursor_keys cursor_keys_mode;
     enum keypad_keys keypad_keys_mode;
     enum mouse_tracking mouse_tracking;
     enum mouse_reporting mouse_reporting;
+    char *mouse_user_cursor;  /* For OSC-22 */
 
     tll(int) tab_stops;
 
@@ -401,6 +406,7 @@ struct terminal {
     struct {
         bool origin:1;
         bool application_cursor_keys:1;
+        bool application_keypad_keys:1;
         bool reverse:1;
         bool show_cursor:1;
         bool reverse_wrap:1;
@@ -422,7 +428,6 @@ struct terminal {
         bool num_lock_modifier:1;
         bool bell_action_enabled:1;
         bool alt_screen:1;
-        bool modify_escape_key:1;
         bool ime:1;
         bool app_sync_updates:1;
 
@@ -484,15 +489,11 @@ struct terminal {
     struct {
         enum selection_kind kind;
         enum selection_direction direction;
-        struct coord start;
-        struct coord end;
+        struct range coords;
         bool ongoing;
         bool spaces_only; /* SELECTION_SEMANTIC_WORD */
 
-        struct {
-            struct coord start;
-            struct coord end;
-        } pivot;
+        struct range pivot;
 
         struct {
             int fd;
@@ -507,7 +508,6 @@ struct terminal {
         size_t len;
         size_t sz;
         size_t cursor;
-        enum { SEARCH_BACKWARD, SEARCH_FORWARD} direction;
 
         int original_view;
         bool view_followed_offset;
@@ -534,6 +534,7 @@ struct terminal {
             struct buffer_chain *render_timer;
             struct buffer_chain *url;
             struct buffer_chain *csd;
+            struct buffer_chain *overlay;
         } chains;
 
         /* Scheduled for rendering, as soon-as-possible */
@@ -587,12 +588,13 @@ struct terminal {
         } last_cursor;
 
         struct buffer *last_buf;     /* Buffer we rendered to last time */
-        bool was_flashing;           /* Flash was active last time we rendered */
-        bool was_searching;
+
+        enum overlay_style last_overlay_style;
+        struct buffer *last_overlay_buf;
+        pixman_region32_t last_overlay_clip;
 
         size_t search_glyph_offset;
 
-        bool presentation_timings;
         struct timespec input_time;
     } render;
 
@@ -663,6 +665,7 @@ struct terminal {
 extern const char *const XCURSOR_HIDDEN;
 extern const char *const XCURSOR_LEFT_PTR;
 extern const char *const XCURSOR_TEXT;
+extern const char *const XCURSOR_TEXT_FALLBACK;
 //extern const char *const XCURSOR_HAND2;
 extern const char *const XCURSOR_TOP_LEFT_CORNER;
 extern const char *const XCURSOR_TOP_RIGHT_CORNER;
@@ -677,7 +680,7 @@ struct config;
 struct terminal *term_init(
     const struct config *conf, struct fdm *fdm, struct reaper *reaper,
     struct wayland *wayl, const char *foot_exe, const char *cwd,
-    const char *token, int argc, char *const *argv,
+    const char *token, int argc, char *const *argv, char *const *envp,
     void (*shutdown_cb)(void *data, int exit_code), void *shutdown_data);
 
 bool term_shutdown(struct terminal *term);
@@ -771,6 +774,7 @@ void term_mouse_motion(
 bool term_mouse_grabbed(const struct terminal *term, const struct seat *seat);
 void term_xcursor_update(struct terminal *term);
 void term_xcursor_update_for_seat(struct terminal *term, struct seat *seat);
+void term_set_user_mouse_cursor(struct terminal *term, const char *cursor);
 
 void term_set_window_title(struct terminal *term, const char *title);
 void term_flash(struct terminal *term, unsigned duration_ms);
