@@ -360,7 +360,7 @@ test_enum(struct context *ctx, bool (*parse_fun)(struct context *ctx),
 
     ctx->value = "invalid-enum-value";
     if (parse_fun(ctx)) {
-        BUG("[%s].%s=%s: did not fail to parse as expeced",
+        BUG("[%s].%s=%s: did not fail to parse as expected",
             ctx->section, ctx->key, ctx->value);
     }
 }
@@ -401,6 +401,52 @@ test_color(struct context *ctx, bool (*parse_fun)(struct context *ctx),
     }
 }
 
+static void
+test_two_colors(struct context *ctx, bool (*parse_fun)(struct context *ctx),
+                const char *key, bool alpha_allowed,
+                uint32_t *ptr1, uint32_t *ptr2)
+{
+    ctx->key = key;
+
+    const struct {
+        const char *option_string;
+        uint32_t color1;
+        uint32_t color2;
+        bool invalid;
+    } input[] = {
+        {"000000 000000", 0, 0},
+
+        /* No alpha */
+        {"999999 888888", 0x999999, 0x888888},
+        {"ffffff aaaaaa", 0xffffff, 0xaaaaaa},
+
+        /* Both colors have alpha component */
+        {"ffffffff 00000000", 0xffffffff, 0x00000000, !alpha_allowed},
+        {"aabbccdd, ee112233", 0xaabbccdd, 0xee112233, !alpha_allowed},
+
+        /* Only one color has alpha component */
+        {"ffffffff 112233", 0xffffffff, 0x112233, !alpha_allowed},
+        {"ffffff ff112233", 0x00ffffff, 0xff112233, !alpha_allowed},
+
+        {"unittest-invalid-color", 0, 0, true},
+    };
+
+    for (size_t i = 0; i < ALEN(input); i++) {
+        ctx->value = input[i].option_string;
+        if (input[i].invalid) {
+            if (parse_fun(ctx)) {
+                BUG("[%s].%s=%s: did not fail to parse as expected",
+                    ctx->section, ctx->key, ctx->value);
+            }
+        } else {
+            if (!parse_fun(ctx)) {
+                BUG("[%s].%s=%s: failed to parse",
+                    ctx->section, ctx->key, ctx->value);
+            }
+        }
+    }
+}
+
 static void
 test_section_main(void)
 {
@@ -464,7 +510,7 @@ test_section_main(void)
     /* TODO: initial-window-size-pixels (geometry) */
     /* TODO: initial-window-size-chars (geometry) */
 
-    config_free(conf);
+    config_free(&conf);
 }
 
 static void
@@ -482,7 +528,7 @@ test_section_bell(void)
     test_spawn_template(&ctx, &parse_section_bell, "command",
                         &conf.bell.command);
 
-    config_free(conf);
+    config_free(&conf);
 }
 
 static void
@@ -509,7 +555,7 @@ test_section_scrollback(void)
 
     /* TODO: indicator-format (enum, sort-of) */
 
-    config_free(conf);
+    config_free(&conf);
 }
 
 static void
@@ -532,7 +578,7 @@ test_section_url(void)
     /* TODO: protocols (list of wchars) */
     /* TODO: uri-characters (wchar string, but sorted) */
 
-    config_free(conf);
+    config_free(&conf);
 }
 
 static void
@@ -558,7 +604,7 @@ test_section_cursor(void)
 
     /* TODO: color (two RRGGBB values) */
 
-    config_free(conf);
+    config_free(&conf);
 }
 
 static void
@@ -575,7 +621,7 @@ test_section_mouse(void)
     test_boolean(&ctx, &parse_section_mouse, "alternate-scroll-mode",
                  &conf.mouse.alternate_scroll_mode);
 
-    config_free(conf);
+    config_free(&conf);
 }
 
 static void
@@ -616,6 +662,18 @@ test_section_colors(void)
     test_color(&ctx, &parse_section_colors, "selection-foreground", false, &conf.colors.selection_fg);
     test_color(&ctx, &parse_section_colors, "selection-background", false, &conf.colors.selection_bg);
     test_color(&ctx, &parse_section_colors, "urls", false, &conf.colors.url);
+    test_two_colors(&ctx, &parse_section_colors, "jump-labels", false,
+                    &conf.colors.jump_label.fg,
+                    &conf.colors.jump_label.bg);
+    test_two_colors(&ctx, &parse_section_colors, "scrollback-indicator", false,
+                    &conf.colors.scrollback_indicator.fg,
+                    &conf.colors.scrollback_indicator.bg);
+    test_two_colors(&ctx, &parse_section_colors, "search-box-no-match", false,
+                    &conf.colors.search_box.no_match.fg,
+                    &conf.colors.search_box.no_match.bg);
+    test_two_colors(&ctx, &parse_section_colors, "search-box-match", false,
+                    &conf.colors.search_box.match.fg,
+                    &conf.colors.search_box.match.bg);
 
     for (size_t i = 0; i < 255; i++) {
         char key_name[4];
@@ -627,10 +685,8 @@ test_section_colors(void)
     test_invalid_key(&ctx, &parse_section_colors, "256");
 
     /* TODO: alpha (float in range 0-1, converted to uint16_t) */
-    /* TODO: jump-labels (two colors) */
-    /* TODO: scrollback-indicator (two colors) */
 
-    config_free(conf);
+    config_free(&conf);
 }
 
 static void
@@ -665,19 +721,21 @@ test_section_csd(void)
     test_color(&ctx, &parse_section_csd, "button-maximize-color", true,
                &conf.csd.color.maximize);
     test_color(&ctx, &parse_section_csd, "button-close-color", true,
-               &conf.csd.color.close);
+               &conf.csd.color.quit);
+    test_boolean(&ctx, &parse_section_csd, "hide-when-maximized",
+                 &conf.csd.hide_when_maximized);
 
     /* TODO: verify the ‘set’ bit is actually set for colors */
     /* TODO: font */
 
-    config_free(conf);
+    config_free(&conf);
 }
 
 static void
 test_key_binding(struct context *ctx, bool (*parse_fun)(struct context *ctx),
                  int action, int max_action, const char *const *map,
                  struct config_key_binding_list *bindings,
-                 enum config_key_binding_type type)
+                 enum key_binding_type type)
 {
     xassert(map[action] != NULL);
     xassert(bindings->count == 0);
@@ -756,30 +814,30 @@ test_key_binding(struct context *ctx, bool (*parse_fun)(struct context *ctx),
         &bindings->arr[bindings->count - 1];
 
     if (argv) {
-        if (binding->pipe.argv.args == NULL) {
+        if (binding->aux.pipe.args == NULL) {
             BUG("[%s].%s=%s: pipe argv is NULL",
                 ctx->section, ctx->key, ctx->value);
         }
 
         for (size_t i = 0; i < ALEN(args); i++) {
-            if (binding->pipe.argv.args[i] == NULL ||
-                strcmp(binding->pipe.argv.args[i], args[i]) != 0)
+            if (binding->aux.pipe.args[i] == NULL ||
+                strcmp(binding->aux.pipe.args[i], args[i]) != 0)
             {
                 BUG("[%s].%s=%s: pipe argv not the expected one: "
                     "mismatch of arg #%zu: expected=\"%s\", got=\"%s\"",
                     ctx->section, ctx->key, ctx->value, i,
-                    args[i], binding->pipe.argv.args[i]);
+                    args[i], binding->aux.pipe.args[i]);
             }
         }
 
-        if (binding->pipe.argv.args[ALEN(args)] != NULL) {
+        if (binding->aux.pipe.args[ALEN(args)] != NULL) {
             BUG("[%s].%s=%s: pipe argv not the expected one: "
                 "expected NULL terminator at arg #%zu, got=\"%s\"",
                 ctx->section, ctx->key, ctx->value,
-                ALEN(args), binding->pipe.argv.args[ALEN(args)]);
+                ALEN(args), binding->aux.pipe.args[ALEN(args)]);
         }
     } else {
-        if (binding->pipe.argv.args != NULL) {
+        if (binding->aux.pipe.args != NULL) {
             BUG("[%s].%s=%s: pipe argv not NULL",
                 ctx->section, ctx->key, ctx->value);
         }
@@ -840,7 +898,7 @@ enum collision_test_mode {
 static void
 _test_binding_collisions(struct context *ctx,
                          int max_action, const char *const *map,
-                         enum config_key_binding_type type,
+                         enum key_binding_type type,
                          enum collision_test_mode test_mode)
 {
     struct config_key_binding *bindings_array =
@@ -889,21 +947,23 @@ _test_binding_collisions(struct context *ctx,
 
     case FAIL_DIFFERENT_ARGV:
     case SUCCED_SAME_ACTION_AND_ARGV:
-        bindings.arr[0].pipe.master_copy = true;
-        bindings.arr[0].pipe.argv.args = xcalloc(
-            4, sizeof(bindings.arr[0].pipe.argv.args[0]));
-        bindings.arr[0].pipe.argv.args[0] = xstrdup("/usr/bin/foobar");
-        bindings.arr[0].pipe.argv.args[1] = xstrdup("hello");
-        bindings.arr[0].pipe.argv.args[2] = xstrdup("world");
+        bindings.arr[0].aux.type = BINDING_AUX_PIPE;
+        bindings.arr[0].aux.master_copy = true;
+        bindings.arr[0].aux.pipe.args = xcalloc(
+            4, sizeof(bindings.arr[0].aux.pipe.args[0]));
+        bindings.arr[0].aux.pipe.args[0] = xstrdup("/usr/bin/foobar");
+        bindings.arr[0].aux.pipe.args[1] = xstrdup("hello");
+        bindings.arr[0].aux.pipe.args[2] = xstrdup("world");
 
-        bindings.arr[1].pipe.master_copy = true;
-        bindings.arr[1].pipe.argv.args = xcalloc(
-            4, sizeof(bindings.arr[1].pipe.argv.args[0]));
-        bindings.arr[1].pipe.argv.args[0] = xstrdup("/usr/bin/foobar");
-        bindings.arr[1].pipe.argv.args[1] = xstrdup("hello");
+        bindings.arr[1].aux.type = BINDING_AUX_PIPE;
+        bindings.arr[1].aux.master_copy = true;
+        bindings.arr[1].aux.pipe.args = xcalloc(
+            4, sizeof(bindings.arr[1].aux.pipe.args[0]));
+        bindings.arr[1].aux.pipe.args[0] = xstrdup("/usr/bin/foobar");
+        bindings.arr[1].aux.pipe.args[1] = xstrdup("hello");
 
         if (test_mode == SUCCED_SAME_ACTION_AND_ARGV)
-            bindings.arr[1].pipe.argv.args[2] = xstrdup("world");
+            bindings.arr[1].aux.pipe.args[2] = xstrdup("world");
         break;
     }
 
@@ -937,7 +997,7 @@ _test_binding_collisions(struct context *ctx,
 static void
 test_binding_collisions(struct context *ctx,
                          int max_action, const char *const *map,
-                        enum config_key_binding_type type)
+                        enum key_binding_type type)
 {
     _test_binding_collisions(ctx, max_action, map, type, FAIL_DIFFERENT_ACTION);
     _test_binding_collisions(ctx, max_action, map, type, FAIL_DIFFERENT_ARGV);
@@ -968,7 +1028,7 @@ test_section_key_bindings(void)
             binding_action_map, &conf.bindings.key, KEY_BINDING);
     }
 
-    config_free(conf);
+    config_free(&conf);
 }
 
 static void
@@ -981,7 +1041,7 @@ test_section_key_bindings_collisions(void)
     test_binding_collisions(
         &ctx, BIND_ACTION_KEY_COUNT - 1, binding_action_map, KEY_BINDING);
 
-    config_free(conf);
+    config_free(&conf);
 }
 
 static void
@@ -1003,7 +1063,7 @@ test_section_search_bindings(void)
             search_binding_action_map, &conf.bindings.search, KEY_BINDING);
     }
 
-    config_free(conf);
+    config_free(&conf);
 }
 
 static void
@@ -1017,7 +1077,7 @@ test_section_search_bindings_collisions(void)
         &ctx,
         BIND_ACTION_SEARCH_COUNT - 1, search_binding_action_map, KEY_BINDING);
 
-    config_free(conf);
+    config_free(&conf);
 }
 
 static void
@@ -1039,7 +1099,7 @@ test_section_url_bindings(void)
             url_binding_action_map, &conf.bindings.url, KEY_BINDING);
     }
 
-    config_free(conf);
+    config_free(&conf);
 }
 
 static void
@@ -1053,7 +1113,7 @@ test_section_url_bindings_collisions(void)
         &ctx,
         BIND_ACTION_URL_COUNT - 1, url_binding_action_map, KEY_BINDING);
 
-    config_free(conf);
+    config_free(&conf);
 }
 
 static void
@@ -1075,7 +1135,7 @@ test_section_mouse_bindings(void)
             binding_action_map, &conf.bindings.mouse, MOUSE_BINDING);
     }
 
-    config_free(conf);
+    config_free(&conf);
 }
 
 static void
@@ -1089,7 +1149,79 @@ test_section_mouse_bindings_collisions(void)
         &ctx,
         BIND_ACTION_COUNT - 1, binding_action_map, MOUSE_BINDING);
 
-    config_free(conf);
+    config_free(&conf);
+}
+
+static void
+test_section_text_bindings(void)
+{
+    struct config conf = {0};
+    struct context ctx = {
+        .conf = &conf, .section = "text-bindings", .path = "unittest"};
+
+    ctx.key = "abcd";
+    ctx.value = XKB_MOD_NAME_CTRL "+" XKB_MOD_NAME_SHIFT "+x";
+    xassert(parse_section_text_bindings(&ctx));
+
+    ctx.key = "\\x07";
+    xassert(parse_section_text_bindings(&ctx));
+
+    ctx.key = "\\x1g";
+    xassert(!parse_section_text_bindings(&ctx));
+
+    ctx.key = "\\x1";
+    xassert(!parse_section_text_bindings(&ctx));
+
+    ctx.key = "\\x";
+    xassert(!parse_section_text_bindings(&ctx));
+
+    ctx.key = "\\";
+    xassert(!parse_section_text_bindings(&ctx));
+
+    ctx.key = "\\y";
+    xassert(!parse_section_text_bindings(&ctx));
+
+    ctx.key = "abcd";
+    ctx.value = "InvalidMod+y";
+    xassert(!parse_section_text_bindings(&ctx));
+
+    config_free(&conf);
+}
+
+static void
+test_section_environment(void)
+{
+    struct config conf = {0};
+    struct context ctx = {
+        .conf = &conf, .section = "environment", .path = "unittest"};
+
+    /* A single variable */
+    ctx.key = "FOO";
+    ctx.value = "bar";
+    xassert(parse_section_environment(&ctx));
+    xassert(tll_length(conf.env_vars) == 1);
+    xassert(strcmp(tll_front(conf.env_vars).name, "FOO") == 0);
+    xassert(strcmp(tll_front(conf.env_vars).value, "bar") == 0);
+
+    /* Add a second variable */
+    ctx.key = "BAR";
+    ctx.value = "123";
+    xassert(parse_section_environment(&ctx));
+    xassert(tll_length(conf.env_vars) == 2);
+    xassert(strcmp(tll_back(conf.env_vars).name, "BAR") == 0);
+    xassert(strcmp(tll_back(conf.env_vars).value, "123") == 0);
+
+    /* Replace the *value* of the first variable */
+    ctx.key = "FOO";
+    ctx.value = "456";
+    xassert(parse_section_environment(&ctx));
+    xassert(tll_length(conf.env_vars) == 2);
+    xassert(strcmp(tll_front(conf.env_vars).name, "FOO") == 0);
+    xassert(strcmp(tll_front(conf.env_vars).value, "456") == 0);
+    xassert(strcmp(tll_back(conf.env_vars).name, "BAR") == 0);
+    xassert(strcmp(tll_back(conf.env_vars).value, "123") == 0);
+
+    config_free(&conf);
 }
 
 static void
@@ -1163,7 +1295,7 @@ test_section_tweak(void)
                 &conf.tweak.max_shm_pool_size);
 #endif
 
-    config_free(conf);
+    config_free(&conf);
 }
 
 int
@@ -1186,6 +1318,8 @@ main(int argc, const char *const *argv)
     test_section_url_bindings_collisions();
     test_section_mouse_bindings();
     test_section_mouse_bindings_collisions();
+    test_section_text_bindings();
+    test_section_environment();
     test_section_tweak();
     log_deinit();
     return 0;
