@@ -13,6 +13,7 @@
 #include "log.h"
 #include "char32.h"
 #include "grid.h"
+#include "key-binding.h"
 #include "render.h"
 #include "selection.h"
 #include "spawn.h"
@@ -25,8 +26,10 @@ static void url_destroy(struct url *url);
 
 static bool
 execute_binding(struct seat *seat, struct terminal *term,
-                enum bind_action_url action, uint32_t serial)
+                const struct key_binding *binding, uint32_t serial)
 {
+    const enum bind_action_url action = binding->action;
+
     switch (action) {
     case BIND_ACTION_URL_NONE:
         return false;
@@ -47,8 +50,86 @@ execute_binding(struct seat *seat, struct terminal *term,
     return true;
 }
 
+static bool
+spawn_url_launcher_with_token(struct terminal *term,
+                              const char *url,
+                              const char *xdg_activation_token)
+{
+    size_t argc;
+    char **argv;
+
+    int dev_null = open("/dev/null", O_RDWR);
+
+    if (dev_null < 0) {
+        LOG_ERRNO("failed to open /dev/null");
+        return false;
+    }
+
+    bool ret = false;
+
+    if (spawn_expand_template(
+            &term->conf->url.launch, 1,
+            (const char *[]){"url"},
+            (const char *[]){url},
+            &argc, &argv))
+    {
+        ret = spawn(term->reaper, term->cwd, argv,
+              dev_null, dev_null, dev_null, xdg_activation_token);
+
+        for (size_t i = 0; i < argc; i++)
+            free(argv[i]);
+        free(argv);
+    }
+
+    close(dev_null);
+    return ret;
+}
+
+#if defined(HAVE_XDG_ACTIVATION)
+struct spawn_activation_context {
+    struct terminal *term;
+    char *url;
+};
+
 static void
-activate_url(struct seat *seat, struct terminal *term, const struct url *url)
+activation_token_done(const char *token, void *data)
+{
+    struct spawn_activation_context *ctx = data;
+
+    spawn_url_launcher_with_token(ctx->term, ctx->url, token);
+    free(ctx->url);
+    free(ctx);
+}
+#endif
+
+static bool
+spawn_url_launcher(struct seat *seat, struct terminal *term, const char *url,
+                   uint32_t serial)
+{
+#if defined(HAVE_XDG_ACTIVATION)
+    struct spawn_activation_context *ctx = xmalloc(sizeof(*ctx));
+    *ctx = (struct spawn_activation_context){
+        .term = term,
+        .url = xstrdup(url),
+    };
+
+    if (wayl_get_activation_token(
+            seat->wayl, seat, serial, term->window, &activation_token_done, ctx))
+    {
+        /* Context free:d by callback */
+        return true;
+    }
+
+    free(ctx->url);
+    free(ctx);
+#endif
+
+    return spawn_url_launcher_with_token(term, url, NULL);
+}
+
+static void
+activate_url(struct seat *seat, struct terminal *term, const struct url *url,
+             uint32_t serial)
 {
     char *url_string = NULL;
 
@@ -82,31 +163,9 @@ activate_url(struct seat *seat, struct terminal *term, const struct url *url)
         }
         break;
 
-    case URL_ACTION_LAUNCH: {
-        size_t argc;
-        char **argv;
-
-        int dev_null = open("/dev/null", O_RDWR);
-
-        if (dev_null < 0) {
-            LOG_ERRNO("failed to open /dev/null");
-            break;
-        }
-
-        if (spawn_expand_template(
-                &term->conf->url.launch, 1,
-                (const char *[]){"url"},
-                (const char *[]){url_string},
-                &argc, &argv))
-        {
-            spawn(term->reaper, term->cwd, argv, dev_null, dev_null, dev_null);
-
-            for (size_t i = 0; i < argc; i++)
-                free(argv[i]);
-            free(argv);
-        }
-
-        close(dev_null);
+    case URL_ACTION_LAUNCH:
+    case URL_ACTION_PERSISTENT: {
+        spawn_url_launcher(seat, term, url_string, serial);
         break;
     }
     }
@@ -115,37 +174,44 @@ activate_url(struct seat *seat, struct terminal *term, const struct url *url)
 }
 
 void
-urls_input(struct seat *seat, struct terminal *term, uint32_t key,
+urls_input(struct seat *seat, struct terminal *term,
+           const struct key_binding_set *bindings, uint32_t key,
            xkb_keysym_t sym, xkb_mod_mask_t mods, xkb_mod_mask_t consumed,
+           xkb_mod_mask_t locked,
            const xkb_keysym_t *raw_syms, size_t raw_count,
            uint32_t serial)
 {
+    const xkb_mod_mask_t bind_mods =
+        mods & seat->kbd.bind_significant & ~locked;
+    const xkb_mod_mask_t bind_consumed =
+        consumed & seat->kbd.bind_significant & ~locked;
+
     /* Key bindings */
-    tll_foreach(seat->kbd.bindings.url, it) {
+    tll_foreach(bindings->url, it) {
         const struct key_binding *bind = &it->item;
 
         /* Match translated symbol */
-        if (bind->sym == sym &&
-            bind->mods == (mods & ~consumed))
+        if (bind->k.sym == sym &&
+            bind->mods == (bind_mods & ~bind_consumed))
         {
-            execute_binding(seat, term, bind->action, serial);
+            execute_binding(seat, term, bind, serial);
             return;
         }
 
-        if (bind->mods != mods)
+        if (bind->mods != bind_mods || bind_mods != (mods & ~locked))
             continue;
 
         for (size_t i = 0; i < raw_count; i++) {
-            if (bind->sym == raw_syms[i]) {
-                execute_binding(seat, term, bind->action, serial);
+            if (bind->k.sym == raw_syms[i]) {
+                execute_binding(seat, term, bind, serial);
                 return;
             }
         }
 
         /* Match raw key code */
-        tll_foreach(bind->key_codes, code) {
+        tll_foreach(bind->k.key_codes, code) {
             if (code->item == key) {
-                execute_binding(seat, term, bind->action, serial);
+                execute_binding(seat, term, bind, serial);
                 return;
             }
         }
@@ -162,13 +228,13 @@ urls_input(struct seat *seat, struct terminal *term, uint32_t key,
         return;
     }
 
-    if (mods & ~consumed)
+    if (mods & ~consumed & ~locked)
         return;
 
     char32_t wc = xkb_state_key_get_utf32(seat->kbd.xkb_state, key);
 
     /*
-     * Determine if this is a “valid” key. I.e. if there is an URL
+     * Determine if this is a “valid” key. I.e. if there is a URL
      * label with a key combo where this key is the next in
      * sequence.
      */
@@ -196,8 +262,19 @@ urls_input(struct seat *seat, struct terminal *term, uint32_t key,
     }
 
     if (match) {
-        activate_url(seat, term, match);
-        urls_reset(term);
+        activate_url(seat, term, match, serial);
+
+        switch (match->action) {
+        case URL_ACTION_COPY:
+        case URL_ACTION_LAUNCH:
+            urls_reset(term);
+            break;
+
+        case URL_ACTION_PERSISTENT:
+            term->url_keys[0] = U'\0';
+            render_refresh_urls(term);
+            break;
+        }
     }
 
     else if (is_valid) {
@@ -403,8 +480,10 @@ auto_detected(const struct terminal *term, enum url_action action,
                             ((struct url){
                                 .id = (uint64_t)rand() << 32 | rand(),
                                 .url = url_utf8,
-                                .start = start,
-                                .end = end,
+                                .range = {
+                                    .start = start,
+                                    .end = end,
+                                },
                                 .action = action,
                                 .osc8 = false}));
                     }
@@ -458,8 +537,10 @@ osc8_uris(const struct terminal *term, enum url_action action, url_list_t *urls)
                ((struct url){
                    .id = range->id,
                    .url = xstrdup(range->uri),
-                   .start = start,
-                   .end = end,
+                   .range = {
+                       .start = start,
+                       .end = end,
+                   },
                    .action = action,
                    .url_mode_dont_change_url_attr = dont_touch_url_attr,
                    .osc8 = true}));
@@ -478,11 +559,11 @@ remove_overlapping(url_list_t *urls, int cols)
             const struct url *out = &outer->item;
             const struct url *in = &inner->item;
 
-            uint64_t in_start = in->start.row * cols + in->start.col;
-            uint64_t in_end = in->end.row * cols + in->end.col;
+            uint64_t in_start = in->range.start.row * cols + in->range.start.col;
+            uint64_t in_end = in->range.end.row * cols + in->range.end.col;
 
-            uint64_t out_start = out->start.row * cols + out->start.col;
-            uint64_t out_end = out->end.row * cols + out->end.col;
+            uint64_t out_start = out->range.start.row * cols + out->range.start.col;
+            uint64_t out_end = out->range.end.row * cols + out->range.end.col;
 
             if ((in_start <= out_start && in_end >= out_start) ||
                 (in_start <= out_end && in_end >= out_end) ||
@@ -596,18 +677,23 @@ urls_assign_key_combos(const struct config *conf, url_list_t *urls)
     if (count == 0)
         return;
 
-    uint64_t seen_ids[count];
     char32_t *combos[count];
     generate_key_combos(conf, count, combos);
 
     size_t combo_idx = 0;
-    size_t id_idx = 0;
 
     tll_foreach(*urls, it) {
         bool id_already_seen = false;
 
-        for (size_t i = 0; i < id_idx; i++) {
-            if (it->item.id == seen_ids[i]) {
+        /* Look for already processed URLs where both the URI and the
+         * ID matches */
+        tll_foreach(*urls, it2) {
+            if (&it->item == &it2->item)
+                break;
+
+            if (it->item.id == it2->item.id &&
+                strcmp(it->item.url, it2->item.url) == 0)
+            {
                 id_already_seen = true;
                 break;
             }
@@ -615,7 +701,6 @@ urls_assign_key_combos(const struct config *conf, url_list_t *urls)
 
         if (id_already_seen)
             continue;
-        seen_ids[id_idx++] = it->item.id;
 
         /*
          * Scan previous URLs, and check if *this* URL matches any of
@@ -649,7 +734,7 @@ urls_assign_key_combos(const struct config *conf, url_list_t *urls)
         char *key = ac32tombs(it->item.key);
         xassert(key != NULL);
 
-        LOG_DBG("URL: %s (%s)", it->item.url, key);
+        LOG_DBG("URL: %s (key=%s, id=%"PRIu64")", it->item.url, key, it->item.id);
         free(key);
     }
 #endif
@@ -661,8 +746,8 @@ tag_cells_for_url(struct terminal *term, const struct url *url, bool value)
     if (url->url_mode_dont_change_url_attr)
         return;
 
-    const struct coord *start = &url->start;
-    const struct coord *end = &url->end;
+    const struct coord *start = &url->range.start;
+    const struct coord *end = &url->range.end;
 
     size_t end_r = end->row & (term->grid->num_rows - 1);
 
@@ -706,7 +791,7 @@ urls_render(struct terminal *term)
     xassert(tll_length(win->urls) == 0);
     tll_foreach(win->term->urls, it) {
         struct wl_url url = {.url = &it->item};
-        wayl_win_subsurface_new(win, &url.surf);
+        wayl_win_subsurface_new(win, &url.surf, false);
 
         tll_push_back(win->urls, url);
         tag_cells_for_url(term, &it->item, true);
