@@ -577,11 +577,20 @@ action_osc_start(struct terminal *term, uint8_t c)
 static void
 action_osc_end(struct terminal *term, uint8_t c)
 {
-    if (!osc_ensure_size(term, term->vt.osc.idx + 1))
+    struct vt *vt = &term->vt;
+
+    if (!osc_ensure_size(term, vt->osc.idx + 1))
         return;
-    term->vt.osc.data[term->vt.osc.idx] = '\0';
-    term->vt.osc.bel = c == '\a';
+
+    vt->osc.data[vt->osc.idx] = '\0';
+    vt->osc.bel = c == '\a';
     osc_dispatch(term);
+
+    if (unlikely(vt->osc.idx >= 4096)) {
+        free(vt->osc.data);
+        vt->osc.data = NULL;
+        vt->osc.size = 0;
+    }
 }
 
 static void
@@ -613,8 +622,21 @@ action_put(struct terminal *term, uint8_t c)
 static inline uint32_t
 chain_key(uint32_t old_key, uint32_t new_wc)
 {
-    /* Rotate left 8 bits, xor with new char */
-    return ((old_key << 8) | (old_key >> (32 - 8))) ^ new_wc;
+    unsigned bits = 32 - __builtin_clz(CELL_COMB_CHARS_HI - CELL_COMB_CHARS_LO);
+
+    /* Rotate old key 8 bits */
+    uint32_t new_key = (old_key << 8) | (old_key >> (bits - 8));
+
+    /* xor with new char */
+    new_key ^= new_wc;
+
+    /* Multiply with magic hash constant */
+    new_key *= 2654435761;
+
+    /* And mask, to ensure the new value is within range */
+    new_key &= CELL_COMB_CHARS_HI - CELL_COMB_CHARS_LO;
+
+    return new_key;
 }
 
 static void
@@ -659,8 +681,6 @@ action_utf8_print(struct terminal *term, char32_t wc)
         } else
             key = chain_key(base, wc);
 
-        key &= CELL_COMB_CHARS_HI - CELL_COMB_CHARS_LO;
-
 #if defined(FOOT_GRAPHEME_CLUSTERING)
         if (grapheme_clustering) {
             /* Check if we're on a grapheme cluster break */
@@ -729,8 +749,20 @@ action_utf8_print(struct terminal *term, char32_t wc)
 
             xassert(wanted_count <= 255);
 
+            size_t collision_count = 0;
+
             /* Look for existing combining chain */
             while (true) {
+                if (unlikely(collision_count > 128)) {
+                    static bool have_warned = false;
+                    if (!have_warned) {
+                        have_warned = true;
+                        LOG_WARN("ignoring composed character: "
+                                 "too many collisions in hash table");
+                    }
+                    return;
+                }
+
                 const struct composed *cc = composed_lookup(term->composed, key);
                 if (cc == NULL)
                     break;
@@ -746,7 +778,12 @@ action_utf8_print(struct terminal *term, char32_t wc)
                     cc->count != wanted_count ||
                     cc->chars[wanted_count - 1] != wc)
                 {
+#if 0
+                    LOG_WARN("COLLISION: base: %04x/%04x, count: %d/%zu, last: %04x/%04x",
+                             cc->chars[0], base, cc->count, wanted_count, cc->chars[wanted_count - 1], wc);
+#endif
                     key++;
+                    collision_count++;
                     continue;
                 }
 
@@ -757,6 +794,7 @@ action_utf8_print(struct terminal *term, char32_t wc)
 
                 if (!match) {
                     key++;
+                    collision_count++;
                     continue;
                 }
 
