@@ -27,6 +27,7 @@
 #include "render.h"
 #include "selection.h"
 #include "shm.h"
+#include "shm-formats.h"
 #include "util.h"
 #include "xmalloc.h"
 
@@ -67,13 +68,14 @@ csd_instantiate(struct wl_window *win)
     xassert(wayl != NULL);
 
     for (size_t i = 0; i < CSD_SURF_MINIMIZE; i++) {
-        bool ret = wayl_win_subsurface_new(win, &win->csd.surface[i]);
+        bool ret = wayl_win_subsurface_new(win, &win->csd.surface[i], true);
         xassert(ret);
     }
 
     for (size_t i = CSD_SURF_MINIMIZE; i < CSD_SURF_COUNT; i++) {
         bool ret = wayl_win_subsurface_new_with_custom_parent(
-            win, win->csd.surface[CSD_SURF_TITLE].surf, &win->csd.surface[i]);
+            win, win->csd.surface[CSD_SURF_TITLE].surf, &win->csd.surface[i],
+            true);
         xassert(ret);
     }
 
@@ -155,12 +157,9 @@ seat_add_text_input(struct seat *seat)
 }
 
 static void
-key_bindings_destroy(key_binding_list_t *bindings)
+seat_add_key_bindings(struct seat *seat)
 {
-    tll_foreach(*bindings, it) {
-        tll_free(it->item.key_codes);
-        tll_remove(*bindings, it);
-    }
+    key_binding_new_for_seat(seat->wayl->key_binding_manager, seat);
 }
 
 static void
@@ -170,12 +169,7 @@ seat_destroy(struct seat *seat)
         return;
 
     tll_free(seat->mouse.buttons);
-
-    key_bindings_destroy(&seat->kbd.bindings.key);
-    key_bindings_destroy(&seat->kbd.bindings.search);
-    key_bindings_destroy(&seat->kbd.bindings.url);
-
-    tll_free(seat->mouse.bindings);
+    key_binding_remove_seat(seat->wayl->key_binding_manager, seat);
 
     if (seat->kbd.xkb_compose_state != NULL)
         xkb_compose_state_unref(seat->kbd.xkb_compose_state);
@@ -234,6 +228,22 @@ static void
 shm_format(void *data, struct wl_shm *wl_shm, uint32_t format)
 {
     struct wayland *wayl = data;
+
+#if defined(_DEBUG)
+    bool have_description = false;
+
+    for (size_t i = 0; i < ALEN(shm_formats); i++) {
+        if (shm_formats[i].format == format) {
+            LOG_DBG("shm: 0x%08x: %s", format, shm_formats[i].description);
+            have_description = true;
+            break;
+        }
+    }
+
+    if (!have_description)
+        LOG_DBG("shm: 0x%08x: unknown", format);
+#endif
+
     if (format == WL_SHM_FORMAT_ARGB8888)
         wayl->have_argb8888 = true;
 }
@@ -696,9 +706,50 @@ xdg_toplevel_close(void *data, struct xdg_toplevel *xdg_toplevel)
     term_shutdown(term);
 }
 
+#if defined(XDG_TOPLEVEL_CONFIGURE_BOUNDS_SINCE_VERSION)
+static void
+xdg_toplevel_configure_bounds(void *data,
+                              struct xdg_toplevel *xdg_toplevel,
+                              int32_t width, int32_t height)
+{
+    /* TODO: ensure we don't pick a bigger size */
+}
+#endif
+
+#if defined(XDG_TOPLEVEL_WM_CAPABILITIES_SINCE_VERSION)
+static void
+xdg_toplevel_wm_capabilities(void *data,
+                             struct xdg_toplevel *xdg_toplevel,
+                             struct wl_array *caps)
+{
+    struct wl_window *win = data;
+
+    win->wm_capabilities.maximize = false;
+    win->wm_capabilities.minimize = false;
+
+    uint32_t *cap_ptr;
+    wl_array_for_each(cap_ptr, caps) {
+        switch (*cap_ptr) {
+        case XDG_TOPLEVEL_WM_CAPABILITIES_MAXIMIZE:
+            win->wm_capabilities.maximize = true;
+            break;
+        case XDG_TOPLEVEL_WM_CAPABILITIES_MINIMIZE:
+            win->wm_capabilities.minimize = true;
+            break;
+        }
+    }
+}
+#endif
+
 static const struct xdg_toplevel_listener xdg_toplevel_listener = {
     .configure = &xdg_toplevel_configure,
     /*.close = */&xdg_toplevel_close,  /* epoll-shim defines a macro ‘close’... */
+#if defined(XDG_TOPLEVEL_CONFIGURE_BOUNDS_SINCE_VERSION)
+    .configure_bounds = &xdg_toplevel_configure_bounds,
+#endif
+#if defined(XDG_TOPLEVEL_WM_CAPABILITIES_SINCE_VERSION)
+    .wm_capabilities = xdg_toplevel_wm_capabilities,
+#endif
 };
 
 static void
@@ -736,8 +787,15 @@ xdg_surface_configure(void *data, struct xdg_surface *xdg_surface,
     else if (csd_was_enabled && !enable_csd)
         csd_destroy(win);
 
-    if (enable_csd && new_width > 0 && new_height > 0)
-        new_height -= win->term->conf->csd.title_height;
+    if (enable_csd && new_width > 0 && new_height > 0) {
+        if (wayl_win_csd_titlebar_visible(win))
+            new_height -= win->term->conf->csd.title_height;
+
+        if (wayl_win_csd_borders_visible(win)) {
+            new_height -= 2 * win->term->conf->csd.border_width_visible;
+            new_width -= 2 * win->term->conf->csd.border_width_visible;
+        }
+    }
 
     xdg_surface_ack_configure(xdg_surface, serial);
 
@@ -879,13 +937,22 @@ handle_global(void *data, struct wl_registry *registry,
             return;
 
         /*
-         * We *require* version 1, but _can_ use version 2. Version 2
+         * We *require* version 1, but _can_ use version 5. Version 2
          * adds 'tiled' window states. We use that information to
-         * restore the window size when window is un-tiled.
+         * restore the window size when window is un-tiled. Version 5
+         * adds 'wm_capabilities'. We use that information to draw
+         * window decorations.
          */
+#if defined(XDG_TOPLEVEL_WM_CAPABILITIES_SINCE_VERSION)
+        const uint32_t preferred = XDG_TOPLEVEL_WM_CAPABILITIES_SINCE_VERSION;
+#elif defined(XDG_TOPLEVEL_STATE_TILED_LEFT_SINCE_VERSION)
+        const uint32_t preferred = XDG_TOPLEVEL_STATE_TILED_LEFT_SINCE_VERSION;
+#else
+        const uint32_t preferred = required;
+#endif
 
         wayl->shell = wl_registry_bind(
-            wayl->registry, name, &xdg_wm_base_interface, min(version, 2));
+            wayl->registry, name, &xdg_wm_base_interface, min(version, preferred));
         xdg_wm_base_add_listener(wayl->shell, &xdg_wm_base_listener, wayl);
     }
 
@@ -934,6 +1001,7 @@ handle_global(void *data, struct wl_registry *registry,
         seat_add_data_device(seat);
         seat_add_primary_selection(seat);
         seat_add_text_input(seat);
+        seat_add_key_bindings(seat);
         wl_seat_add_listener(wl_seat, &seat_listener, seat);
     }
 
@@ -1012,7 +1080,7 @@ handle_global(void *data, struct wl_registry *registry,
     }
 
     else if (strcmp(interface, wp_presentation_interface.name) == 0) {
-        if (wayl->conf->presentation_timings) {
+        if (wayl->presentation_timings) {
             const uint32_t required = 1;
             if (!verify_iface_version(interface, version, required))
                 return;
@@ -1189,7 +1257,8 @@ fdm_wayl(struct fdm *fdm, int fd, int events, void *data)
 }
 
 struct wayland *
-wayl_init(const struct config *conf, struct fdm *fdm)
+wayl_init(struct fdm *fdm, struct key_binding_manager *key_binding_manager,
+          bool presentation_timings)
 {
     struct wayland *wayl = calloc(1, sizeof(*wayl));
     if (unlikely(wayl == NULL)) {
@@ -1197,9 +1266,10 @@ wayl_init(const struct config *conf, struct fdm *fdm)
         return NULL;
     }
 
-    wayl->conf = conf;
     wayl->fdm = fdm;
+    wayl->key_binding_manager = key_binding_manager;
     wayl->fd = -1;
+    wayl->presentation_timings = presentation_timings;
 
     if (!fdm_hook_add(fdm, &fdm_hook, wayl, FDM_HOOK_PRIORITY_LOW)) {
         LOG_ERR("failed to add FDM hook");
@@ -1250,16 +1320,16 @@ wayl_init(const struct config *conf, struct fdm *fdm)
         LOG_WARN("no primary selection available");
 
 #if defined(HAVE_XDG_ACTIVATION)
-    if (wayl->xdg_activation == NULL && conf->bell.urgent) {
+    if (wayl->xdg_activation == NULL) {
 #else
-    if (conf->bell.urgent) {
+    if (true) {
 #endif
         LOG_WARN(
             "no XDG activation support; "
             "bell.urgent will fall back to coloring the window margins red");
     }
 
-    if (conf->presentation_timings && wayl->presentation == NULL) {
+    if (presentation_timings && wayl->presentation == NULL) {
         LOG_ERR("presentation time interface not implemented by compositor");
         goto out;
     }
@@ -1398,6 +1468,9 @@ wayl_win_init(struct terminal *term, const char *token)
     win->csd.move_timeout_fd = -1;
     win->resize_timeout_fd = -1;
 
+    win->wm_capabilities.maximize = true;
+    win->wm_capabilities.minimize = true;
+
     win->surface = wl_compositor_create_surface(wayl->compositor);
     if (win->surface == NULL) {
         LOG_ERR("failed to create wayland surface");
@@ -1458,10 +1531,15 @@ wayl_win_init(struct terminal *term, const char *token)
         xdg_activation_v1_activate(wayl->xdg_activation, token, win->surface);
 #endif
 
+    if (!wayl_win_subsurface_new(win, &win->overlay, false)) {
+        LOG_ERR("failed to create overlay surface");
+        goto out;
+    }
+
     switch (conf->tweak.render_timer) {
     case RENDER_TIMER_OSD:
     case RENDER_TIMER_BOTH:
-        if (!wayl_win_subsurface_new(win, &win->render_timer)) {
+        if (!wayl_win_subsurface_new(win, &win->render_timer, false)) {
             LOG_ERR("failed to create render timer surface");
             goto out;
         }
@@ -1548,6 +1626,7 @@ wayl_win_destroy(struct wl_window *win)
     wayl_win_subsurface_destroy(&win->search);
     wayl_win_subsurface_destroy(&win->scrollback_indicator);
     wayl_win_subsurface_destroy(&win->render_timer);
+    wayl_win_subsurface_destroy(&win->overlay);
 
     shm_purge(term->render.chains.search);
     shm_purge(term->render.chains.scrollback_indicator);
@@ -1557,8 +1636,12 @@ wayl_win_destroy(struct wl_window *win)
     shm_purge(term->render.chains.csd);
 
 #if defined(HAVE_XDG_ACTIVATION)
-    if (win->xdg_activation_token != NULL)
-        xdg_activation_token_v1_destroy(win->xdg_activation_token);
+    tll_foreach(win->xdg_tokens, it) {
+        xdg_activation_token_v1_destroy(it->item->xdg_token);
+        free(it->item);
+
+        tll_remove(win->xdg_tokens, it);
+    }
 #endif
     if (win->frame_callback != NULL)
         wl_callback_destroy(win->frame_callback);
@@ -1697,60 +1780,58 @@ wayl_roundtrip(struct wayland *wayl)
 
 #if defined(HAVE_XDG_ACTIVATION)
 static void
-activation_token_done(void *data, struct xdg_activation_token_v1 *xdg_token,
-                      const char *token)
+activation_token_for_urgency_done(const char *token, void *data)
 {
     struct wl_window *win = data;
     struct wayland *wayl = win->term->wl;
 
-    LOG_DBG("activation token: %s", token);
-
+    win->urgency_token_is_pending = false;
     xdg_activation_v1_activate(wayl->xdg_activation, token, win->surface);
-
-    xassert(win->xdg_activation_token == xdg_token);
-    xdg_activation_token_v1_destroy(xdg_token);
-    win->xdg_activation_token = NULL;
 }
-
-static const struct xdg_activation_token_v1_listener activation_token_listener = {
-    .done = &activation_token_done,
-};
 #endif /* HAVE_XDG_ACTIVATION */
 
 bool
 wayl_win_set_urgent(struct wl_window *win)
 {
 #if defined(HAVE_XDG_ACTIVATION)
-    struct wayland *wayl = win->term->wl;
-
-    if (wayl->xdg_activation == NULL)
-        return false;
-
-    if (win->xdg_activation_token != NULL)
+    if (win->urgency_token_is_pending) {
+        /* We already have a pending token. Don’t request another one,
+         * to avoid flooding the Wayland socket */
         return true;
-
-    struct xdg_activation_token_v1 *token =
-        xdg_activation_v1_get_activation_token(wayl->xdg_activation);
-
-    if (token == NULL) {
-        LOG_ERR("failed to retrieve XDG activation token");
-        return false;
     }
 
-    xdg_activation_token_v1_add_listener(token, &activation_token_listener, win);
-    xdg_activation_token_v1_set_surface(token, win->surface);
-    xdg_activation_token_v1_commit(token);
-    win->xdg_activation_token = token;
-    return true;
-#else
-    return false;
+    bool success = wayl_get_activation_token(
+        win->term->wl, NULL, 0, win, &activation_token_for_urgency_done, win);
+
+    if (success) {
+        win->urgency_token_is_pending = true;
+        return true;
+    }
 #endif
+
+    return false;
+}
+
+bool
+wayl_win_csd_titlebar_visible(const struct wl_window *win)
+{
+    return win->csd_mode == CSD_YES &&
+        !win->is_fullscreen &&
+        !(win->is_maximized && win->term->conf->csd.hide_when_maximized);
+}
+
+bool
+wayl_win_csd_borders_visible(const struct wl_window *win)
+{
+    return win->csd_mode == CSD_YES &&
+        !win->is_fullscreen &&
+        !win->is_maximized;
 }
 
 bool
 wayl_win_subsurface_new_with_custom_parent(
     struct wl_window *win, struct wl_surface *parent,
-    struct wl_surf_subsurf *surf)
+    struct wl_surf_subsurf *surf, bool allow_pointer_input)
 {
     struct wayland *wayl = win->term->wl;
 
@@ -1774,15 +1855,25 @@ wayl_win_subsurface_new_with_custom_parent(
     wl_surface_set_user_data(main_surface, win);
     wl_subsurface_set_sync(sub);
 
+    /* Disable pointer and touch events */
+    if (!allow_pointer_input) {
+        struct wl_region *empty =
+            wl_compositor_create_region(wayl->compositor);
+        wl_surface_set_input_region(main_surface, empty);
+        wl_region_destroy(empty);
+    }
+
     surf->surf = main_surface;
     surf->sub = sub;
     return true;
 }
 
 bool
-wayl_win_subsurface_new(struct wl_window *win, struct wl_surf_subsurf *surf)
+wayl_win_subsurface_new(struct wl_window *win, struct wl_surf_subsurf *surf,
+                        bool allow_pointer_input)
 {
-    return wayl_win_subsurface_new_with_custom_parent(win, win->surface, surf);
+    return wayl_win_subsurface_new_with_custom_parent(
+        win, win->surface, surf, allow_pointer_input);
 }
 
 void
@@ -1798,3 +1889,72 @@ wayl_win_subsurface_destroy(struct wl_surf_subsurf *surf)
     surf->surf = NULL;
     surf->sub = NULL;
 }
+
+#if defined(HAVE_XDG_ACTIVATION)
+
+static void
+activation_token_done(void *data, struct xdg_activation_token_v1 *xdg_token,
+                      const char *token)
+{
+    LOG_DBG("XDG activation token done: %s", token);
+
+    struct xdg_activation_token_context *ctx = data;
+    struct wl_window *win = ctx->win;
+
+    ctx->cb(token, ctx->cb_data);
+
+    tll_foreach(win->xdg_tokens, it) {
+        if (it->item->xdg_token != xdg_token)
+            continue;
+
+        xassert(win == it->item->win);
+
+        free(ctx);
+        xdg_activation_token_v1_destroy(xdg_token);
+        tll_remove(win->xdg_tokens, it);
+        return;
+    }
+
+    xassert(false);
+}
+
+static const struct
+xdg_activation_token_v1_listener activation_token_listener = {
+    .done = &activation_token_done,
+};
+
+bool
+wayl_get_activation_token(
+    struct wayland *wayl, struct seat *seat, uint32_t serial,
+    struct wl_window *win,
+    void (*cb)(const char *token, void *data), void *cb_data)
+{
+    if (wayl->xdg_activation == NULL)
+        return false;
+
+    struct xdg_activation_token_v1 *token =
+        xdg_activation_v1_get_activation_token(wayl->xdg_activation);
+
+    if (token == NULL) {
+        LOG_ERR("failed to retrieve XDG activation token");
+        return false;
+    }
+
+    struct xdg_activation_token_context *ctx = xmalloc(sizeof(*ctx));
+    *ctx = (struct xdg_activation_token_context){
+        .win = win,
+        .xdg_token = token,
+        .cb = cb,
+        .cb_data = cb_data,
+    };
+    tll_push_back(win->xdg_tokens, ctx);
+
+    if (seat != NULL && serial != 0)
+        xdg_activation_token_v1_set_serial(token, serial, seat->wl_seat);
+
+    xdg_activation_token_v1_set_surface(token, win->surface);
+    xdg_activation_token_v1_add_listener(token, &activation_token_listener, ctx);
+    xdg_activation_token_v1_commit(token);
+    return true;
+}
+#endif
